"""Testing utilities for building expected queries for review requests.

Version Added:
    5.0.7
"""

from __future__ import annotations

from typing import Dict, Optional, Sequence, Set, TYPE_CHECKING, Union

from django.contrib.auth.models import AnonymousUser, User
from django.db.models import OuterRef, Q, Value

from reviewboard.accounts.testing.queries import get_user_profile_equeries
from reviewboard.reviews.models import Group, ReviewRequest
from reviewboard.reviews.testing.queries.review_groups import \
    get_review_groups_accessible_ids_equeries
from reviewboard.scmtools.models import Repository
from reviewboard.scmtools.testing.queries import \
    get_repositories_accessible_ids_equeries
from reviewboard.site.models import LocalSite

if TYPE_CHECKING:
    from django_assert_queries.query_comparator import ExpectedQueries
    from typing_extensions import NotRequired, TypedDict, Unpack

    from reviewboard.accounts.models import Profile
    from reviewboard.site.models import AnyOrAllLocalSites
    from reviewboard.testing.queries.base import ExpectedQResult

    class _AccessibleKwargs(TypedDict):
        user: Optional[Union[AnonymousUser, User]]
        local_site: NotRequired[AnyOrAllLocalSites]
        has_local_sites_in_db: NotRequired[bool]
        show_all_unpublished: NotRequired[bool]
        show_inactive: NotRequired[bool]
        filter_private: NotRequired[bool]
        status: NotRequired[Optional[str]]
        extra_query: NotRequired[Q]
        accessible_repository_ids: NotRequired[Sequence[int]]
        accessible_review_group_ids: NotRequired[Sequence[int]]
        needs_local_site_profile_query: NotRequired[bool]
        needs_user_permission_queries: NotRequired[bool]


##########################
# Q-expression utilities #
##########################

def get_review_requests_accessible_q(
    *,
    user: Optional[Union[AnonymousUser, User]],
    local_site: AnyOrAllLocalSites = None,
    has_local_sites_in_db: bool = False,
    show_all_unpublished: bool = False,
    show_inactive: bool = False,
    filter_private: bool = False,
    status: Optional[str] = 'P',
    extra_query: Q = Q(),
    accessible_repository_ids: Sequence[int] = [],
    accessible_review_group_ids: Sequence[int] = [],
    needs_local_site_profile_query: bool = False,
    needs_user_permission_queries: bool = True,
) -> ExpectedQResult:
    """Return a Q expression for accessible review request queries.

    Version Added:
        5.0.7

    Args:
        user (django.contrib.auth.models.AnonymousUser or
              django.contrib.auth.models.User):
            The user that's checked for access.

        local_site (reviewboard.site.models.LocalSite, optional):
            The optional Local Site value used in the queries.

        has_local_sites_in_db (bool, optional):
            Whether the query expects Local Sites to be in the database.

        show_all_unpublished (bool, optional):
            Whether unpublished review requests would be allowed in the
            results.

        show_inactive (bool, optional):
            Whether review requests from inactive users would be allowed in
            the results.

        filter_private (bool, optional):
            Whether review requests would be filtered using ACL checks.

        status (str, optional):
            The optional status that would be used in the query.

        extra_query (django.db.models.Q, optional):
            An extra query filter that would be included.

        accessible_repository_ids (list of int, optional):
            A list of accessible repository IDs that would be expected in the
            query.

        accessible_review_group_ids (list of int, optional):
            A list of accessible review group IDs that would be expected in the
            query.

        needs_local_site_profile_query (bool, optional):
            Whether the query should fetch a
            :py:class:`~reviewboard.accounts.models.LocalSiteProfile`.

            Set this to ``False`` if this should be cached at this point.

        needs_user_permission_queries (bool, optional):
            Whether the query should check for the
            ``reviews.can_view_invite_only_groups`` permission.

            Set to ``False`` if this is already cached at this point.

    Returns:
        dict:
        The expected Q results.
    """
    include_local_site_q = (has_local_sites_in_db and
                            local_site is not LocalSite.ALL)

    prep_equeries: ExpectedQueries = []
    subqueries: ExpectedQueries = []

    is_authenticated: bool
    is_superuser: bool

    if user is None or isinstance(user, AnonymousUser):
        is_authenticated = False
        is_superuser = False
    else:
        is_authenticated = True
        is_superuser = user.is_superuser

        if filter_private:
            prep_equeries = get_review_requests_accessible_prep_equeries(
                user=user,
                local_site=local_site,
                needs_local_site_profile_query=needs_local_site_profile_query,
                needs_user_permission_queries=needs_user_permission_queries)

    # This is intended to be verbose, to ensure we're matching exactly the
    # queries we expect. We want to minimize building of queries.
    #
    # That said, to keep this maintainable, query is built in 2 stages:
    #
    # 1. Initial caller-requested access control/visibility queries
    # 2. Access control filter queries
    init_q: Q
    access_q: Q
    tables: Set[str] = {'reviews_reviewrequest'}
    join_types: Dict[str, str] = {}

    # Stage 1 has the following states (excluding Local Sites in DB):
    #
    # +----+---------------+----------------------+---------------+--------+
    # |  Q | authenticated | show_all_unpublished | show_inactive | status |
    # +----+---------------+----------------------+---------------+--------+
    # |  1 | X             | X                    | X             | x      |
    # |  2 | X             | X                    | X             |        |
    # |  3 | X             | X                    |               | X      |
    # |  4 | X             | X                    |               |        |
    # |  5 | X             |                      | X             | X      |
    # |  6 | X             |                      | X             |        |
    # |  7 | X             |                      |               | X      |
    # |  8 | X             |                      |               |        |
    # |  9 |               | X                    | X             | x      |
    # | 10 |               | X                    | X             |        |
    # | 11 |               | X                    |               | X      |
    # | 12 |               | X                    |               |        |
    # | 13 |               |                      | X             | X      |
    # | 14 |               |                      | X             |        |
    # | 15 |               |                      |               | X      |
    # | 16 |               |                      |               |        |
    # +----+---------------+----------------------+---------------+--------+
    #
    # Some of the resulting queries are identical, but will be repeated.
    if is_authenticated:
        if show_all_unpublished:
            if show_inactive and status:
                # Q 1
                if include_local_site_q:
                    init_q = (Q(status=status) &
                              Q(local_site=local_site) &
                              extra_query)
                else:
                    init_q = (Q(status=status) &
                              extra_query)
            elif show_inactive and not status:
                # Q 2
                if include_local_site_q:
                    init_q = (Q(local_site=local_site) &
                              extra_query)
                else:
                    init_q = extra_query
            elif not show_inactive and status:
                # Q 3
                tables.add('auth_user')
                join_types['auth_user'] = 'INNER JOIN'

                if include_local_site_q:
                    init_q = (Q(submitter__is_active=True) &
                              Q(status=status) &
                              Q(local_site=local_site) &
                              extra_query)
                else:
                    init_q = (Q(submitter__is_active=True) &
                              Q(status=status) &
                              extra_query)
            else:
                # Q 4
                assert not show_inactive
                assert not status

                tables.add('auth_user')
                join_types['auth_user'] = 'INNER JOIN'

                if include_local_site_q:
                    init_q = (Q(submitter__is_active=True) &
                              Q(local_site=local_site) &
                              extra_query)
                else:
                    init_q = (Q(submitter__is_active=True) &
                              extra_query)
        else:
            # show_all_unpublished=False

            if show_inactive and status:
                # Q 5
                if include_local_site_q:
                    init_q = ((Q(public=True) |
                               Q(submitter=user)) &
                              Q(local_site=local_site) &
                              extra_query)
                else:
                    init_q = ((Q(public=True) |
                               Q(submitter=user)) &
                              Q(status=status) &
                              extra_query)
            elif show_inactive and not status:
                # Q 6
                if include_local_site_q:
                    init_q = ((Q(public=True) |
                               Q(submitter=user)) &
                              Q(local_site=local_site) &
                              extra_query)
                else:
                    init_q = ((Q(public=True) |
                               Q(submitter=user)) &
                              extra_query)
            elif not show_inactive and status:
                # Q 7
                tables.add('auth_user')
                join_types['auth_user'] = 'INNER JOIN'

                if include_local_site_q:
                    init_q = ((Q(public=True) |
                               Q(submitter=user)) &
                              Q(submitter__is_active=True) &
                              Q(status=status) &
                              Q(local_site=local_site) &
                              extra_query)
                else:
                    init_q = ((Q(public=True) |
                               Q(submitter=user)) &
                              Q(submitter__is_active=True) &
                              Q(status=status) &
                              extra_query)
            else:
                # Q 8
                assert not show_inactive
                assert not status

                tables.add('auth_user')
                join_types['auth_user'] = 'INNER JOIN'

                if include_local_site_q:
                    init_q = ((Q(public=True) |
                               Q(submitter=user)) &
                              Q(submitter__is_active=True) &
                              Q(local_site=local_site) &
                              extra_query)
                else:
                    init_q = ((Q(public=True) |
                               Q(submitter=user)) &
                              Q(submitter__is_active=True) &
                              extra_query)
    else:
        # Anonymous.
        if show_all_unpublished:
            if show_inactive and status:
                # Q 9
                if include_local_site_q:
                    init_q = (Q(status=status) &
                              Q(local_site=local_site) &
                              extra_query)
                else:
                    init_q = (Q(status=status) &
                              extra_query)
            elif show_inactive and not status:
                # Q 10
                if include_local_site_q:
                    init_q = (Q(local_site=local_site) &
                              extra_query)
                else:
                    init_q = extra_query
            elif not show_inactive and status:
                # Q 11
                tables.add('auth_user')
                join_types['auth_user'] = 'INNER JOIN'

                if include_local_site_q:
                    init_q = (Q(submitter__is_active=True) &
                              Q(status=status) &
                              Q(local_site=local_site) &
                              extra_query)
                else:
                    init_q = (Q(submitter__is_active=True) &
                              Q(status=status) &
                              extra_query)
            else:
                # Q 12
                assert not show_inactive
                assert not status

                tables.add('auth_user')
                join_types['auth_user'] = 'INNER JOIN'

                if include_local_site_q:
                    init_q = (Q(submitter__is_active=True) &
                              Q(local_site=local_site) &
                              extra_query)
                else:
                    init_q = (Q(submitter__is_active=True) &
                              extra_query)
        else:
            # show_all_unpublished=False

            if show_inactive and status:
                # Q 13
                if include_local_site_q:
                    init_q = (Q(public=True) &
                              Q(local_site=local_site) &
                              extra_query)
                else:
                    init_q = (Q(public=True) &
                              Q(status=status) &
                              extra_query)
            elif show_inactive and not status:
                # Q 14
                if include_local_site_q:
                    init_q = (Q(public=True) &
                              Q(local_site=local_site) &
                              extra_query)
                else:
                    init_q = (Q(public=True) &
                              extra_query)
            elif not show_inactive and status:
                # Q 15
                tables.add('auth_user')
                join_types['auth_user'] = 'INNER JOIN'

                if include_local_site_q:
                    init_q = (Q(public=True) &
                              Q(submitter__is_active=True) &
                              Q(status=status) &
                              Q(local_site=local_site) &
                              extra_query)
                else:
                    init_q = (Q(public=True) &
                              Q(submitter__is_active=True) &
                              Q(status=status) &
                              extra_query)
            else:
                # Q 16
                assert not show_inactive
                assert not status

                tables.add('auth_user')
                join_types['auth_user'] = 'INNER JOIN'

                if include_local_site_q:
                    init_q = (Q(public=True) &
                              Q(submitter__is_active=True) &
                              Q(local_site=local_site) &
                              extra_query)
                else:
                    init_q = (Q(public=True) &
                              Q(submitter__is_active=True) &
                              extra_query)

    # Stage 2: Access control filters.
    if filter_private:
        if is_superuser:
            access_q = Q()
        else:
            if is_authenticated:
                access_q = (Q(submitter=user) |
                            (Q(repository=None) |
                             Q(repository__in=accessible_repository_ids)) &
                            (Q(__Exists__subquery__=1) |
                             Q(
                                 ~Q(__Exists__subquery__=2) |
                                 Q(__Exists__subquery__=3))))

                subqueries += [
                    {
                        'model': ReviewRequest.target_people.through,
                        'annotations': {
                            'a': Value(1),
                        },
                        'limit': 1,
                        'subquery': True,
                        'where': (Q(reviewrequest_id=OuterRef('pk')) &
                                  Q(user=user)),
                    },
                    {
                        'model': ReviewRequest.target_groups.through,
                        'annotations': {
                            'a': Value(1),
                        },
                        'limit': 1,
                        'subquery': True,
                        'where': Q(reviewrequest_id=OuterRef('pk')),
                    },
                    {
                        'model': ReviewRequest.target_groups.through,
                        'annotations': {
                            'a': Value(1),
                        },
                        'limit': 1,
                        'subquery': True,
                        'where': (Q(reviewrequest_id=OuterRef('pk')) &
                                  Q(group__in=accessible_review_group_ids)),
                    },
                ]
            else:
                access_q = ((Q(repository=None) |
                             Q(__Exists__subquery__=1)) &
                            (~Q(__Exists__subquery__=2) |
                             Q(__Exists__subquery__=3)))

                subqueries += [
                    {
                        'model': Repository,
                        'annotations': {
                            'a': Value(1),
                        },
                        'limit': 1,
                        'subquery': True,
                        'where': (Q(pk=OuterRef('repository_id')) &
                                  Q(public=True)),
                    },
                    {
                        'model': ReviewRequest.target_groups.through,
                        'annotations': {
                            'a': Value(1),
                        },
                        'limit': 1,
                        'subquery': True,
                        'where': Q(reviewrequest_id=OuterRef('pk')),
                    },
                    {
                        'join_types': {
                            'reviews_group': 'INNER JOIN',
                        },
                        'model': ReviewRequest.target_groups.through,
                        'annotations': {
                            'a': Value(1),
                        },
                        'limit': 1,
                        'subquery': True,
                        'num_joins': 1,
                        'tables': {
                            'reviews_group',
                            'reviews_reviewrequest_target_groups',
                        },
                        'where': (Q(reviewrequest_id=OuterRef('pk')) &
                                  Q(group__invite_only=False)),
                    },
                ]
    else:
        access_q = Q()

    return {
        'join_types': join_types,
        'prep_equeries': prep_equeries,
        'q': init_q & access_q,
        'subqueries': subqueries,
        'tables': tables,
    }


def get_review_requests_from_user_q(
    *,
    from_user: Union[str, User],
    **kwargs: Unpack[_AccessibleKwargs],
) -> ExpectedQResult:
    """Return a Q expression for review requests from a user.

    This corresponds to a call from
    :py:meth:`ReviewRequest.objects.from_user()
    <reviewboard.reviews.managers.ReviewRequestManager.from_user>`.

    It wraps :py:func:`get_review_requests_accessible_q` and takes all the same
    arguments.

    Version Added:
        5.0.7

    Args:
        from_user (str or django.contrib.auth.models.User):
            The username or user that review requests would be published from
            for the query.

        **kwargs (dict):
            Keyword arguments to pass to
            :py:func:`get_review_requests_accessible_q`.

    Returns:
        dict:
        The expected Q results.
    """
    extra_tables: Set[str] = set()
    extra_join_types: Dict[str, str] = {}

    if isinstance(from_user, User):
        extra_query = Q(submitter=from_user)
    else:
        assert isinstance(from_user, str)

        extra_query = Q(submitter__username=from_user)
        extra_tables.add('auth_user')
        extra_join_types['auth_user'] = 'INNER JOIN'

    kwargs['extra_query'] = extra_query

    result_q = get_review_requests_accessible_q(**kwargs)
    result_q['tables'].update(extra_tables)
    result_q.setdefault('join_types', {}).update(extra_join_types)

    return result_q


def get_review_requests_to_group_q(
    *,
    to_group_name: str,
    **kwargs: Unpack[_AccessibleKwargs],
) -> ExpectedQResult:
    """Return a Q expression for review requests to a group.

    This corresponds to a call from
    :py:meth:`ReviewRequest.objects.to_group()
    <reviewboard.reviews.managers.ReviewRequestManager.to_group>`.

    It wraps :py:func:`get_review_requests_accessible_q` and takes all the same
    arguments.

    Version Added:
        5.0.7

    Args:
        to_group_name (str):
            The name of the group.

        **kwargs (dict):
            Keyword arguments to pass to
            :py:func:`get_review_requests_accessible_q`.

    Returns:
        dict:
        The expected Q results.
    """
    has_local_sites_in_db = kwargs.get('has_local_sites_in_db', False)

    if has_local_sites_in_db:
        extra_query = (Q(target_groups__name=to_group_name) &
                       Q(target_groups__local_site=kwargs.get('local_site')))
    else:
        extra_query = Q(target_groups__name=to_group_name)

    kwargs['extra_query'] = extra_query

    q_result = get_review_requests_accessible_q(**kwargs)
    q_result['tables'].update({
        'reviews_group',
        'reviews_reviewrequest_target_groups',
    })
    q_result.setdefault('join_types', {}).update({
        'reviews_group': 'INNER JOIN',
        'reviews_reviewrequest_target_groups': 'INNER JOIN',
    })

    return q_result


def get_review_requests_to_user_q(
    *,
    to_user: Union[str, User],
    to_user_profile: Optional[Profile],
    target_groups: Sequence[Group] = [],
    **kwargs: Unpack[_AccessibleKwargs],
) -> ExpectedQResult:
    """Return a Q expression for review requests to a user.

    This corresponds to a call from
    :py:meth:`ReviewRequest.objects.to_user()
    <reviewboard.reviews.managers.ReviewRequestManager.to_user>`.

    It wraps :py:func:`get_review_requests_accessible_q` and takes all the same
    arguments.

    The ``prep_equeries`` key of the result must immediately precede any
    Usage of the resulting Q-expression.

    Version Added:
        5.0.7

    Args:
        to_user (str or django.contrib.auth.models.User):
            The username or user that review requests would be assigned to.

        to_user_profile (reviewboard.accounts.models.Profile):
            The profile of the user that review requests would be assigned to.

        target_groups (list of reviewboard.reviews.models.group.Group):
            The list of review groups that the user may be a member of,
            for the query.

        **kwargs (dict):
            Keyword arguments to pass to
            :py:func:`get_review_requests_accessible_q`.

    Returns:
        dict:
        The expected Q results.
    """
    extra_join_types: Dict[str, str] = {}
    extra_tables: Set[str] = set()

    target_group_ids = [
        _group.pk
        for _group in target_groups
    ]

    prep_equeries: ExpectedQueries = []

    if isinstance(to_user, User):
        to_username = to_user.username
        needs_profile = False
    else:
        assert isinstance(to_user, str)

        to_username = to_user
        to_user = (
            User.objects
            .filter(username=to_user)
            .only('pk')
            .get()
        )
        needs_profile = True

        prep_equeries += [
            {
                '__note__': f'Fetching user {to_username}',
                'model': User,
                'where': Q(username=to_username),
            },
        ]

    if to_user_profile is None:
        extra_query = (Q(__Exists__subquery__=1) |
                       Q(__Exists__subquery__=2))
    else:
        extra_query = (Q(__Exists__subquery__=1) |
                       Q(__Exists__subquery__=2) |
                       Q(starred_by=to_user_profile))
        extra_tables.add('accounts_profile_starred_review_requests')
        extra_join_types['accounts_profile_starred_review_requests'] = \
            'LEFT OUTER JOIN'

    kwargs['extra_query'] = extra_query

    q_result = get_review_requests_accessible_q(**kwargs)

    prep_equeries += q_result.get('prep_equeries', []) + [
        {
            '__note__': (
                f'Fetch the list of user "{to_username}"\'s review groups'
            ),
            'join_types': {
                'reviews_group_users': 'INNER JOIN',
            },
            'model': Group,
            'num_joins': 1,
            'values_select': ('pk',),
            'tables': {
                'reviews_group',
                'reviews_group_users',
            },
            'where': Q(users__id=to_user.pk),
        },
    ]

    if needs_profile:
        prep_equeries += get_user_profile_equeries(user=to_user)

    q_result['tables'].update(extra_tables)
    q_result['prep_equeries'] = prep_equeries
    q_result.setdefault('join_types', {}).update(extra_join_types)
    q_result.setdefault('subqueries', []).extend([
        {
            'model': ReviewRequest.target_people.through,
            'annotations': {
                'a': Value(1),
            },
            'limit': 1,
            'subquery': True,
            'where': (Q(reviewrequest_id=OuterRef('pk')) &
                      Q(user=to_user)),
        },
        {
            'model': ReviewRequest.target_groups.through,
            'annotations': {
                'a': Value(1),
            },
            'limit': 1,
            'subquery': True,
            'where': (Q(reviewrequest_id=OuterRef('pk')) &
                      Q(group__in=target_group_ids)),
        }
    ])

    return q_result


def get_review_requests_to_user_directly_q(
    *,
    to_user: Union[str, User],
    to_user_profile: Optional[Profile],
    subquery_start_index: int = 1,
    **kwargs: Unpack[_AccessibleKwargs],
) -> ExpectedQResult:
    """Return a Q expression for review requests directly to a user.

    This corresponds to a call from
    :py:meth:`ReviewRequest.objects.to_user_directly()
    <reviewboard.reviews.managers.ReviewRequestManager.to_user_directly>`.

    It wraps :py:func:`get_review_requests_accessible_q` and takes all the same
    arguments.

    Version Added:
        5.0.7

    Args:
        to_user (str or django.contrib.auth.models.User):
            The username or user that review requests would be assigned to.

        to_user_profile (reviewboard.accounts.models.Profile):
            The profile of the user that review requests would be assigned to.

        subquery_start_index (int, optional):
            The starting index for any subquery markers, to ease embedding
            in another query.

        **kwargs (dict):
            Keyword arguments to pass to
            :py:func:`get_review_requests_accessible_q`.

    Returns:
        dict:
        The expected Q results.
    """
    extra_join_types: Dict[str, str] = {}
    extra_tables: Set[str] = set()
    prep_equeries: ExpectedQueries = []

    if isinstance(to_user, User):
        to_username = to_user.username
        needs_profile = False
    else:
        assert isinstance(to_user, str)

        to_username = to_user
        to_user = (
            User.objects
            .filter(username=to_user)
            .only('pk')
            .get()
        )
        needs_profile = True

        prep_equeries += [
            {
                '__note__': f'Fetching user {to_username}',
                'model': User,
                'where': Q(username=to_username),
            },
        ]

    if to_user_profile is None:
        extra_query = Q(__Exists__subquery__=subquery_start_index)
    else:
        extra_query = (Q(__Exists__subquery__=subquery_start_index) |
                       Q(starred_by=to_user_profile))
        extra_tables.add('accounts_profile_starred_review_requests')
        extra_join_types['accounts_profile_starred_review_requests'] = \
            'LEFT OUTER JOIN'

    kwargs['extra_query'] = extra_query

    q_result = get_review_requests_accessible_q(**kwargs)

    prep_equeries += q_result.get('prep_equeries', [])

    if needs_profile:
        prep_equeries += get_user_profile_equeries(user=to_user)

    q_result['prep_equeries'] = prep_equeries
    q_result['tables'].update(extra_tables)
    q_result.setdefault('join_types', {}).update(extra_join_types)
    q_result.setdefault('subqueries', []).append({
        'model': ReviewRequest.target_people.through,
        'annotations': {
            'a': Value(1),
        },
        'limit': 1,
        'subquery': True,
        'where': (Q(reviewrequest_id=OuterRef('pk')) &
                  Q(user=to_user)),
    })

    return q_result


def get_review_requests_to_user_groups_q(
    *,
    to_user: Union[str, User],
    to_user_profile: Optional[Profile],
    target_groups: Sequence[Group] = [],
    **kwargs: Unpack[_AccessibleKwargs],
) -> ExpectedQResult:
    """Return a Q expression for review requests to a user's groups.

    This corresponds to a call from
    :py:meth:`ReviewRequest.objects.to_user_groups()
    <reviewboard.reviews.managers.ReviewRequestManager.to_user_groups>`.

    It wraps :py:func:`get_review_requests_accessible_q` and takes all the same
    arguments.

    The ``prep_equeries`` key of the result must immediately precede any
    Usage of the resulting Q-expression.

    Version Added:
        5.0.7

    Args:
        to_user (str or django.contrib.auth.models.User):
            The username or user belonging to a group that review requests
            would be assigned to.

        to_user_profile (reviewboard.accounts.models.Profile):
            The profile of the user belonging to a group that review requests
            would be assigned to.

        target_groups (list of reviewboard.reviews.models.group.Group):
            The list of review groups that the user may be a member of,
            for the query.

        **kwargs (dict):
            Keyword arguments to pass to
            :py:func:`get_review_requests_accessible_q`.

    Returns:
        dict:
        The expected Q results.
    """
    prep_equeries: ExpectedQueries = []

    if isinstance(to_user, User):
        to_username = to_user.username
    else:
        assert isinstance(to_user, str)

        to_username = to_user
        to_user = (
            User.objects
            .filter(username=to_user)
            .only('pk')
            .get()
        )

        prep_equeries += [
            {
                '__note__': f'Fetching user {to_username}',
                'model': User,
                'where': Q(username=to_username),
            },
        ]

    kwargs['extra_query'] = Q(target_groups__in=[
        _group.pk
        for _group in target_groups
    ])

    q_result = get_review_requests_accessible_q(**kwargs)

    prep_equeries += q_result.get('prep_equeries', []) + [
        {
            '__note__': (
                f'Fetch the list of user "{to_username}"\'s review groups'
            ),
            'join_types': {
                'reviews_group_users': 'INNER JOIN',
            },
            'model': Group,
            'num_joins': 1,
            'values_select': ('pk',),
            'tables': {
                'reviews_group',
                'reviews_group_users',
            },
            'where': Q(users__id=to_user.pk),
        },
    ]

    q_result['tables'].add('reviews_reviewrequest_target_groups')
    q_result['prep_equeries'] = prep_equeries
    q_result.setdefault('join_types', {}).update({
        'reviews_reviewrequest_target_groups': 'INNER JOIN',
    })

    return q_result


def get_review_requests_to_or_from_user_q(
    *,
    to_or_from_user: Union[str, User],
    to_or_from_user_profile: Optional[Profile],
    target_groups: Sequence[Group] = [],
    **kwargs: Unpack[_AccessibleKwargs],
) -> ExpectedQResult:
    """Return a Q expression for review requests to/from a user.

    This corresponds to a call from
    :py:meth:`ReviewRequest.objects.to_or_from_user()
    <reviewboard.reviews.managers.ReviewRequestManager.to_or_from_user>`.

    It wraps :py:func:`get_review_requests_accessible_q` and takes all the same
    arguments.

    The ``prep_equeries`` key of the result must immediately precede any
    Usage of the resulting Q-expression.

    Version Added:
        5.0.7

    Args:
        to_or_from_user (str or django.contrib.auth.models.User):
            The username or user that review requests would be assigned to
            or posted from.

        to_or_from_user_profile (reviewboard.accounts.models.Profile):
            The profile of the user that review requests would be assigned to
            or posted from.

        target_groups (list of reviewboard.reviews.models.group.Group):
            The list of review groups that the user may be a member of,
            for the query.

        **kwargs (dict):
            Keyword arguments to pass to
            :py:func:`get_review_requests_accessible_q`.

    Returns:
        dict:
        The expected Q results.
    """
    extra_tables: Set[str] = set()
    extra_join_types: Dict[str, str] = {}

    target_group_ids = [
        _group.pk
        for _group in target_groups
    ]

    prep_equeries: ExpectedQueries = []

    if isinstance(to_or_from_user, User):
        to_or_from_username = to_or_from_user.username
        has_user_obj = True
        needs_profile = False
    else:
        assert isinstance(to_or_from_user, str)

        to_or_from_username = to_or_from_user
        to_or_from_user = (
            User.objects
            .filter(username=to_or_from_user)
            .only('pk')
            .get()
        )

        has_user_obj = False
        needs_profile = True

        prep_equeries += [
            {
                '__note__': f'Fetching user {to_or_from_username}',
                'model': User,
                'where': Q(username=to_or_from_username),
            },
        ]

    if to_or_from_user_profile is None:
        if has_user_obj:
            extra_query = (Q(__Exists__subquery__=1) |
                           Q(__Exists__subquery__=2) |
                           Q(submitter=to_or_from_user))
        else:
            extra_tables.add('auth_user')
            extra_join_types['auth_user'] = 'INNER JOIN'
            extra_query = (Q(__Exists__subquery__=1) |
                           Q(__Exists__subquery__=2) |
                           Q(submitter__username=to_or_from_username))
    else:
        extra_tables.add('accounts_profile_starred_review_requests')
        extra_join_types['accounts_profile_starred_review_requests'] = \
            'LEFT OUTER JOIN'

        if has_user_obj:
            extra_query = (Q(__Exists__subquery__=1) |
                           Q(__Exists__subquery__=2) |
                           Q(starred_by=to_or_from_user_profile) |
                           Q(submitter=to_or_from_user))
        else:
            extra_tables.add('auth_user')
            extra_join_types['auth_user'] = 'INNER JOIN'
            extra_query = (Q(__Exists__subquery__=1) |
                           Q(__Exists__subquery__=2) |
                           Q(starred_by=to_or_from_user_profile) |
                           Q(submitter__username=to_or_from_username))

    kwargs['extra_query'] = extra_query

    q_result = get_review_requests_accessible_q(**kwargs)

    prep_equeries += q_result.get('prep_equeries', []) + [
        {
            '__note__': (
                f'Fetch the list of user "{to_or_from_username}"\'s review '
                f'groups'
            ),
            'join_types': {
                'reviews_group_users': 'INNER JOIN',
            },
            'model': Group,
            'num_joins': 1,
            'values_select': ('pk',),
            'tables': {
                'reviews_group',
                'reviews_group_users',
            },
            'where': Q(users__id=to_or_from_user.pk),
        },
    ]

    if needs_profile:
        prep_equeries += get_user_profile_equeries(user=to_or_from_user)

    q_result['tables'].update(extra_tables)
    q_result['prep_equeries'] = prep_equeries
    q_result.setdefault('join_types', {}).update(extra_join_types)
    q_result.setdefault('subqueries', []).extend([
        {
            'model': ReviewRequest.target_people.through,
            'annotations': {
                'a': Value(1),
            },
            'limit': 1,
            'subquery': True,
            'where': (Q(reviewrequest_id=OuterRef('pk')) &
                      Q(user=to_or_from_user)),
        },
        {
            'model': ReviewRequest.target_groups.through,
            'annotations': {
                'a': Value(1),
            },
            'limit': 1,
            'subquery': True,
            'where': (Q(reviewrequest_id=OuterRef('pk')) &
                      Q(group__in=target_group_ids)),
        }
    ])

    return q_result


############################
# Expected Query utilities #
############################

def get_review_requests_accessible_prep_equeries(
    *,
    user: Union[AnonymousUser, User],
    local_site: AnyOrAllLocalSites = None,
    needs_local_site_profile_query: bool = True,
    needs_user_permission_queries: bool = True,
) -> ExpectedQueries:
    """Return expected review request accessibility preparation queries.

    These are queries that must be performed before the main accessibility
    query is executed.

    Version Added:
        5.0.7

    Args:
        user (django.contrib.auth.models.AnonymousUser or
              django.contrib.auth.models.User):
            The user that's checked for access.

        local_site (reviewboard.site.models.LocalSite, optional):
            The optional Local Site value used in the queries.

        needs_local_site_profile_query (bool, optional):
            Whether the query should fetch a
            :py:class:`~reviewboard.accounts.models.LocalSiteProfile`.

            Set to ``False`` if this should be cached at this point.

        needs_user_permission_queries (bool, optional):
            Whether the query should check for the
            ``reviews.can_view_invite_only_groups`` permission.

            Set to ``False`` if this is already cached at this point.

    Returns:
        list of dict:
        The list of expected queries.
    """
    equeries: ExpectedQueries = []

    if user.is_authenticated and not user.is_superuser:
        equeries += get_repositories_accessible_ids_equeries(
            user=user,
            local_site=local_site,
            visible_only=False)
        equeries += get_review_groups_accessible_ids_equeries(
            user=user,
            local_site=local_site,
            visible_only=False,
            needs_local_site_profile_query=needs_local_site_profile_query,
            needs_user_permission_queries=needs_user_permission_queries)

    return equeries


def get_review_requests_accessible_equeries(
    *,
    distinct: bool = False,
    **kwargs: Unpack[_AccessibleKwargs],
) -> ExpectedQueries:
    """Return queries for accessible review requests.

    This corresponds to a call from
    :py:meth:`ReviewRequest.objects.public()
    <reviewboard.reviews.managers.ReviewRequestManager.public>`.

    Version Added:
        5.0.7

    Args:
        distinct (bool, optional):
            Whether this is a distinct query.

        **kwargs (dict, optional):
            Keyword arguments to pass to
            :py:func:`get_review_requests_accessible_q`.

    Returns:
        list of dict:
        The list of expected queries.
    """
    kwargs['filter_private'] = True

    q_result = get_review_requests_accessible_q(**kwargs)
    q_tables = q_result['tables']

    equeries = q_result.get('prep_equeries', [])
    equeries += [
        {
            '__note__': 'Fetch accessible review requests',
            'distinct': distinct,
            'join_types': q_result.get('join_types', {}),
            'num_joins': len(q_tables) - 1,
            'model': ReviewRequest,
            'subqueries': q_result.get('subqueries', []),
            'tables': q_tables,
            'where': q_result['q'],
        }
    ]

    return equeries


def get_review_requests_from_user_equeries(
    *,
    from_user: Union[str, User],
    distinct: bool = False,
    **kwargs: Unpack[_AccessibleKwargs],
) -> ExpectedQueries:
    """Return queries for review requests from a user.

    This corresponds to a call from
    :py:meth:`ReviewRequest.objects.from_user()
    <reviewboard.reviews.managers.ReviewRequestManager.from_user>`.

    Version Added:
        5.0.7

    Args:
        from_user (str or django.contrib.auth.models.User):
            The username or user that review requests would be published from
            for the query.

        distinct (bool, optional):
            Whether this is a distinct query.

        **kwargs (dict, optional):
            Keyword arguments to pass to
            :py:func:`get_review_requests_accessible_q`.

    Returns:
        list of dict:
        The list of expected queries.
    """
    q_result = get_review_requests_from_user_q(from_user=from_user,
                                               **kwargs)

    if isinstance(from_user, User):
        username = from_user.username
    else:
        username = from_user

    equeries = q_result.get('prep_equeries', [])
    equeries += [
        {
            '__note__': f'Fetch review requests from user "{username}"',
            'distinct': distinct,
            'join_types': q_result.get('join_types', {}),
            'num_joins': 1,
            'model': ReviewRequest,
            'subqueries': q_result.get('subqueries', []),
            'tables': {
                'auth_user',
                'reviews_reviewrequest',
            },
            'where': q_result['q'],
        }
    ]

    return equeries


def get_review_requests_to_group_equeries(
    *,
    to_group_name: str,
    distinct: bool = False,
    **kwargs: Unpack[_AccessibleKwargs],
) -> ExpectedQueries:
    """Return queries for review requests to a group.

    This corresponds to a call from
    :py:meth:`ReviewRequest.objects.to_group()
    <reviewboard.reviews.managers.ReviewRequestManager.to_group>`.

    Version Added:
        5.0.7

    Args:
        to_group (str):
            The name of the group that review requests would be assigned to.

        distinct (bool, optional):
            Whether this is a distinct query.

        **kwargs (dict, optional):
            Keyword arguments to pass to
            :py:func:`get_review_requests_accessible_q`.

    Returns:
        list of dict:
        The list of expected queries.
    """
    q_result = get_review_requests_to_group_q(to_group_name=to_group_name,
                                              **kwargs)
    q_tables = q_result['tables']

    equeries = q_result.get('prep_equeries', [])
    equeries += [
        {
            '__note__': f'Fetch review requests to group "{to_group_name}"',
            'distinct': distinct,
            'join_types': q_result.get('join_types', {}),
            'num_joins': len(q_tables) - 1,
            'model': ReviewRequest,
            'subqueries': q_result.get('subqueries', []),
            'tables': q_tables,
            'where': q_result['q'],
        }
    ]

    return equeries


def get_review_requests_to_user_equeries(
    *,
    to_user: Union[str, User],
    to_user_profile: Optional[Profile] = None,
    target_groups: Sequence[Group] = [],
    distinct: bool = False,
    **kwargs: Unpack[_AccessibleKwargs],
) -> ExpectedQueries:
    """Return queries for review requests to a user.

    This corresponds to a call from
    :py:meth:`ReviewRequest.objects.to_user()
    <reviewboard.reviews.managers.ReviewRequestManager.to_user>`.

    Version Added:
        5.0.7

    Args:
        to_user (str or django.contrib.auth.models.User):
            The username or user that review requests would be assigned to.

        to_user_profile (reviewboard.accounts.models.Profile):
            The profile of the user that review requests would be assigned to.

        target_groups (list of reviewboard.reviews.models.group.Group):
            The list of review groups that the user may be a member of,
            for the query.

        distinct (bool, optional):
            Whether this is a distinct query.

        **kwargs (dict, optional):
            Keyword arguments to pass to
            :py:func:`get_review_requests_accessible_q`.

    Returns:
        list of dict:
        The list of expected queries.
    """
    q_result = get_review_requests_to_user_q(
        to_user=to_user,
        to_user_profile=to_user_profile,
        target_groups=target_groups,
        **kwargs)
    q_tables = q_result['tables']

    if isinstance(to_user, User):
        username = to_user.username
    else:
        username = to_user

    equeries = q_result.get('prep_equeries', [])
    equeries += [
        {
            '__note__': f'Fetch review requests to user "{username}"',
            'distinct': distinct,
            'join_types': q_result.get('join_types', {}),
            'num_joins': len(q_tables) - 1,
            'model': ReviewRequest,
            'subqueries': q_result.get('subqueries', []),
            'tables': q_tables,
            'where': q_result['q'],
        }
    ]

    return equeries


def get_review_requests_to_user_directly_equeries(
    *,
    to_user: Union[str, User],
    to_user_profile: Optional[Profile] = None,
    distinct: bool = False,
    **kwargs: Unpack[_AccessibleKwargs],
) -> ExpectedQueries:
    """Return queries for review requests directly to a user.

    This corresponds to a call from
    :py:meth:`ReviewRequest.objects.to_user_directly()
    <reviewboard.reviews.managers.ReviewRequestManager.to_user_directly>`.

    Version Added:
        5.0.7

    Args:
        to_user (str or django.contrib.auth.models.User):
            The username or user that review requests would be assigned to.

        to_user_profile (reviewboard.accounts.models.Profile):
            The profile of the user that review requests would be assigned to.

        target_groups (list of reviewboard.reviews.models.group.Group):
            The list of review groups that the user may be a member of,
            for the query.

        distinct (bool, optional):
            Whether this is a distinct query.

        **kwargs (dict, optional):
            Keyword arguments to pass to
            :py:func:`get_review_requests_accessible_q`.

    Returns:
        list of dict:
        The list of expected queries.
    """
    q_result = get_review_requests_to_user_directly_q(
        to_user=to_user,
        to_user_profile=to_user_profile,
        **kwargs)
    q_tables = q_result['tables']

    if isinstance(to_user, User):
        username = to_user.username
    else:
        username = to_user

    equeries = q_result.get('prep_equeries', [])
    equeries += [
        {
            '__note__': f'Fetch review requests to user "{username}"',
            'distinct': distinct,
            'join_types': q_result.get('join_types', {}),
            'num_joins': len(q_tables) - 1,
            'model': ReviewRequest,
            'subqueries': q_result.get('subqueries', []),
            'tables': q_tables,
            'where': q_result['q'],
        }
    ]

    return equeries


def get_review_requests_to_user_groups_equeries(
    *,
    to_user: Union[str, User],
    to_user_profile: Optional[Profile] = None,
    target_groups: Sequence[Group] = [],
    distinct: bool = False,
    **kwargs: Unpack[_AccessibleKwargs],
) -> ExpectedQueries:
    """Return queries for review requests to a user's groups.

    This corresponds to a call from
    :py:meth:`ReviewRequest.objects.to_user_groups()
    <reviewboard.reviews.managers.ReviewRequestManager.to_user_groups>`.

    Version Added:
        5.0.7

    Args:
        to_user (str or django.contrib.auth.models.User):
            The username or user belonging to a group that review requests
            would be assigned to.

        to_user_profile (reviewboard.accounts.models.Profile):
            The profile of the user belonging to a group that review requests
            would be assigned to.

        target_groups (list of reviewboard.reviews.models.group.Group):
            The list of review groups that the user may be a member of,
            for the query.

        distinct (bool, optional):
            Whether this is a distinct query.

        **kwargs (dict, optional):
            Keyword arguments to pass to
            :py:func:`get_review_requests_accessible_q`.

    Returns:
        list of dict:
        The list of expected queries.
    """
    q_result = get_review_requests_to_user_groups_q(
        to_user=to_user,
        to_user_profile=to_user_profile,
        target_groups=target_groups,
        **kwargs)
    q_tables = q_result['tables']

    if isinstance(to_user, User):
        username = to_user.username
    else:
        username = to_user

    equeries = q_result.get('prep_equeries', [])
    equeries += [
        {
            '__note__': f'Fetch review requests to user "{username}"',
            'distinct': distinct,
            'join_types': q_result.get('join_types', {}),
            'num_joins': len(q_tables) - 1,
            'model': ReviewRequest,
            'subqueries': q_result.get('subqueries', []),
            'tables': q_tables,
            'where': q_result['q'],
        }
    ]

    return equeries


def get_review_requests_to_or_from_user_equeries(
    *,
    to_or_from_user: Union[str, User],
    to_or_from_user_profile: Optional[Profile] = None,
    target_groups: Sequence[Group] = [],
    distinct: bool = False,
    **kwargs: Unpack[_AccessibleKwargs],
) -> ExpectedQueries:
    """Return queries for review requests to/from a user.

    This corresponds to a call from
    :py:meth:`ReviewRequest.objects.to_user()
    <reviewboard.reviews.managers.ReviewRequestManager.to_user>`.

    Version Added:
        5.0.7

    Args:
        to_or_from_user (str or django.contrib.auth.models.User):
            The username or user that review requests would be assigned to
            or posted from.

        to_or_from_user_profile (reviewboard.accounts.models.Profile):
            The profile of the user that review requests would be assigned to
            or posted from.

        target_groups (list of reviewboard.reviews.models.group.Group):
            The list of review groups that the user may be a member of,
            for the query.

        distinct (bool, optional):
            Whether this is a distinct query.

        **kwargs (dict, optional):
            Keyword arguments to pass to
            :py:func:`get_review_requests_accessible_q`.

    Returns:
        list of dict:
        The list of expected queries.
    """
    q_result = get_review_requests_to_or_from_user_q(
        to_or_from_user=to_or_from_user,
        to_or_from_user_profile=to_or_from_user_profile,
        target_groups=target_groups,
        **kwargs)
    q_tables = q_result['tables']

    if isinstance(to_or_from_user, User):
        username = to_or_from_user.username
    else:
        username = to_or_from_user

    equeries = q_result.get('prep_equeries', [])
    equeries += [
        {
            '__note__': f'Fetch review requests to user "{username}"',
            'distinct': distinct,
            'join_types': q_result.get('join_types', {}),
            'num_joins': len(q_tables) - 1,
            'model': ReviewRequest,
            'subqueries': q_result.get('subqueries', []),
            'tables': q_tables,
            'where': q_result['q'],
        }
    ]

    return equeries



"""Base class for screenshot comment resources."""

from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Q
from django.utils.timesince import timesince
from djblets.util.decorators import augment_method_from
from djblets.webapi.fields import (IntFieldType,
                                   ResourceFieldType,
                                   StringFieldType)

from reviewboard.reviews.models import ScreenshotComment
from reviewboard.webapi.base import WebAPIResource
from reviewboard.webapi.resources import resources
from reviewboard.webapi.resources.base_comment import BaseCommentResource


class BaseScreenshotCommentResource(BaseCommentResource):
    """Base class for screenshot comment resources.

    Provides common fields and functionality for all screenshot comment
    resources. The list of comments cannot be modified from this resource.
    """

    model = ScreenshotComment
    name = 'screenshot_comment'

    fields = dict({
        'screenshot': {
            'type': ResourceFieldType,
            'resource': 'reviewboard.webapi.resources.screenshot.'
                        'ScreenshotResource',
            'description': 'The screenshot the comment was made on.',
        },
        'x': {
            'type': IntFieldType,
            'description': 'The X location of the comment region on the '
                           'screenshot.',
        },
        'y': {
            'type': IntFieldType,
            'description': 'The Y location of the comment region on the '
                           'screenshot.',
        },
        'w': {
            'type': IntFieldType,
            'description': 'The width of the comment region on the '
                           'screenshot.',
        },
        'h': {
            'type': IntFieldType,
            'description': 'The height of the comment region on the '
                           'screenshot.',
        },
        'thumbnail_url': {
            'type': StringFieldType,
            'description': 'The URL to an image showing what was commented '
                           'on.',
            'added_in': '1.7.10',
        },
    }, **BaseCommentResource.fields)

    uri_object_key = 'comment_id'

    allowed_methods = ('GET',)

    def get_queryset(self, request, review_request_id=None, *args, **kwargs):
        """Return a queryset for ScreenshotComment models.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            review_request_id (int, optional):
                The review request ID used to filter the results. If set,
                only comments from the given review request that are public
                or owned by the requesting user will be included.

            *args (tuple):
                Additional positional arguments.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            django.db.models.query.QuerySet:
            A queryset for ScreenshotComment models.
        """
        q = Q(review__isnull=False)

        if review_request_id is not None:
            try:
                review_request = resources.review_request.get_object(
                    request, review_request_id=review_request_id,
                    *args, **kwargs)
            except ObjectDoesNotExist:
                raise self.model.DoesNotExist

            q &= (Q(screenshot__review_request=review_request) |
                  Q(screenshot__inactive_review_request=review_request))

        return self.model.objects.filter(q)

    def serialize_public_field(self, obj, **kwargs):
        return obj.review.get().public

    def serialize_timesince_field(self, obj, **kwargs):
        return timesince(obj.timestamp)

    def serialize_user_field(self, obj, **kwargs):
        return obj.review.get().user

    def serialize_thumbnail_url_field(self, obj, **kwargs):
        return obj.get_image_url()

    @augment_method_from(WebAPIResource)
    def get(self, *args, **kwargs):
        """Returns information on the comment.

        This contains the comment text, time the comment was made,
        and the location of the comment region on the screenshot, amongst
        other information. It can be used to reconstruct the exact
        position of the comment for use as an overlay on the screenshot.
        """
        pass



"""Storage backend support for SSL/TLS certificates.

Version Added:
    6.0
"""

from __future__ import annotations

from typing import cast

from djblets.registries.importer import lazy_import_registry

from reviewboard.certs.storage.registry import \
    CertificateStorageBackendRegistry


#: The registry managing available storage backends.
#:
#: Version Added:
#:     6.0
#:
#: Type:
#:     reviewboard.certs.storage.registry.CertificateStorageBackendRegistry
#:
cert_storage_backend_registry = cast(
    CertificateStorageBackendRegistry,
    lazy_import_registry('reviewboard.certs.storage.registry',
                         'CertificateStorageBackendRegistry'))



"""Increase max_length of FileAttachment.file from 100 to 512.

Version Added:
    1.6.19
"""

from django_evolution.mutations import ChangeField


MUTATIONS = [
    ChangeField('FileAttachment', 'file', initial=None, max_length=512)
]



"""Resources representing commits on a multi-commit review request draft."""

import logging

from djblets.util.decorators import augment_method_from
from djblets.webapi.decorators import (webapi_login_required,
                                       webapi_request_fields,
                                       webapi_response_errors)
from djblets.webapi.errors import (DOES_NOT_EXIST, INVALID_ATTRIBUTE,
                                   INVALID_FORM_DATA)
from djblets.webapi.fields import (DateTimeFieldType,
                                   FileFieldType,
                                   StringFieldType)

from reviewboard.diffviewer.errors import DiffTooBigError, EmptyDiffError
from reviewboard.reviews.forms import UploadCommitForm
from reviewboard.reviews.models import ReviewRequest, ReviewRequestDraft
from reviewboard.scmtools.core import FileNotFoundError
from reviewboard.webapi.decorators import webapi_check_local_site
from reviewboard.webapi.errors import (DIFF_EMPTY,
                                       DIFF_TOO_BIG,
                                       REPO_FILE_NOT_FOUND)
from reviewboard.webapi.resources import resources
from reviewboard.webapi.resources.diffcommit import DiffCommitResource


logger = logging.getLogger(__name__)


class DraftDiffCommitResource(DiffCommitResource):
    """Provides information on pending draft commits for a review request.

    POSTing to this resource will update a review request draft with the
    provided diff.
    """

    name = 'draft_commit'
    model_parent_key = 'diffset'

    allowed_methods = ('GET', 'POST', 'PUT')

    def get_queryset(self, request, *args, **kwargs):
        """Return a QuerySet limited to the review request draft.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            *args (tuple):
                Additional positional arguments.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            django.db.models.query.QuerySet:
            The generated QuerySet.
        """
        try:
            draft = resources.review_request_draft.get_object(request, *args,
                                                              **kwargs)
        except ReviewRequestDraft.DoesNotExist:
            return self.model.objects.none()

        if draft.diffset_id is None:
            return self.model.objects.none()

        return self.model.objects.filter(diffset__pk=draft.diffset_id)

    def has_access_permissions(self, request, commit, *args, **kwargs):
        """Return whether or not the user has access permissions to the commit.

        A user has access permissions for a commit if they have permission to
        access the review request draft.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            commit (reviewboard.diffviewer.models.diffcommit.DiffCommit):
                The object to check access permissions for.

            *args (tuple):
                Additional positional arguments.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            bool:
            Whether or not the user has permission to access the commit.
        """
        draft = resources.review_request_draft.get_object(request, *args,
                                                          **kwargs)
        return draft.is_accessible_by(request.user)

    def has_list_access_permissions(self, request, *args, **kwargs):
        """Return whether the user has access permissions to the list resource.

        A user has list access permissions if they have permission to access
        the review request draft.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            commit (reviewboard.diffviewer.models.diffcommit.DiffCommit):
                The object to check access permissions for.

            *args (tuple):
                Additional positional arguments.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            bool:
            Whether or not the user has permission to access the list resource.
        """
        draft = resources.review_request_draft.get_object(request, *args,
                                                          **kwargs)
        return draft.is_accessible_by(request.user)

    def has_modify_permissions(self, request, commit, *args, **kwargs):
        """Return whether the user has access permissions to modify the object.

        A user has modify permissions for a commit if they have permission to
        modify the review request draft.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            commit (reviewboard.diffviewer.models.diffcommit.DiffCommit):
                The object to check access permissions for.

            *args (tuple):
                Additional positional arguments.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            bool:
            Whether or not the user has permission to modify the object.
        """
        draft = resources.review_request_draft.get_object(request, *args,
                                                          **kwargs)
        return draft.is_mutable_by(request.user)

    @webapi_login_required
    @augment_method_from(DiffCommitResource)
    def get(self, *args, **kwargs):
        pass

    @webapi_login_required
    @augment_method_from(DiffCommitResource)
    def get_list(self, *args, **kwargs):
        pass

    @webapi_login_required
    @webapi_check_local_site
    @webapi_response_errors(DIFF_EMPTY, DIFF_TOO_BIG, DOES_NOT_EXIST,
                            INVALID_ATTRIBUTE, INVALID_FORM_DATA,
                            REPO_FILE_NOT_FOUND)
    @webapi_request_fields(
        required={
            'diff': {
                'type': FileFieldType,
                'description': 'The corresponding diff for this commit.',
            },
            'commit_id': {
                'type': StringFieldType,
                'description': 'The ID of this commit.',
            },
            'author_name': {
                'type': StringFieldType,
                'description': 'The name of the author of this commit.',
            },
            'author_date': {
                'type': DateTimeFieldType,
                'description': 'The date and time this commit was authored in '
                               'ISO 8601 format (YYYY-MM-DD HH:MM:SS+ZZZZ).',
            },
            'author_email': {
                'type': StringFieldType,
                'description': 'The e-mail address of the author of this '
                               'commit.',
            },
            'commit_message': {
                'type': StringFieldType,
                'description': 'The commit message.',
            },
        },
        optional={
            'committer_name': {
                'type': StringFieldType,
                'description': (
                    'The name of the the committer of this commit, if '
                    'applicable.\n'
                    '\n'
                    'If this field is specified, the "committer_date" and '
                    '"committer_email" fields must also be specified.'
                ),
            },
            'committer_date': {
                'type': StringFieldType,
                'description': (
                    'The date and time this commit was committed in ISO 8601 '
                    'format (YYYY-MM-DD HH:MM:SS+ZZZZ).\n'
                    '\n'
                    'If this field is specified, the "committer_name" and '
                    '"committer_email" fields must also be specified.'
                ),
            },
            'committer_email': {
                'type': StringFieldType,
                'description': (
                    'The e-mail address of the committer of this commit.\n'
                    '\n'
                    'If this field is specified, the "committer_name" and '
                    '"committer_date" fields must also be specified.'
                ),
            },
            'parent_diff': {
                'type': FileFieldType,
                'description': 'The optional parent diff to upload.',
            },
            'validation_info': {
                'type': StringFieldType,
                'description': (
                    'Validation metadata from the :ref:`DiffCommit validation '
                    'resource <webapi2.0-validate-diff-commit-resource>`.'
                    '\n\n'
                    'This is required for all but the first commit.'
                ),
            },
        },
        allow_unknown=True
    )
    def create(self, request, extra_fields=None, *args, **kwargs):
        """Create a new commit.

        A draft must exist and the review request must be created with history
        support in order to post to this resource.
        """
        try:
            review_request = resources.review_request.get_object(
                request, *args, **kwargs)
        except ReviewRequest.DoesNotExist:
            return DOES_NOT_EXIST

        if not self.has_modify_permissions(request, review_request, *args,
                                           **kwargs):
            return self.get_no_access_error(request)

        if review_request.repository is None:
            return INVALID_ATTRIBUTE, {
                'reason': 'This review request was created as attachments-'
                          'only, with no repository.',
            }
        elif not review_request.created_with_history:
            reverse_kwargs = {
                'review_request_id': kwargs['review_request_id'],
            }

            if request.local_site:
                reverse_kwargs['local_site_name'] = request.local_site.name

            return INVALID_ATTRIBUTE, {
                'reason': (
                    'This review request was not created with support for '
                    'multiple commits.\n\n'
                    'Use the %(name)s resource to upload diffs instead. See '
                    'the %(name)s link on the parent resource for the URL.'
                    % {
                        'name': resources.draft_diff.name,
                    }
                ),
            }

        draft = review_request.get_draft(user=request.user)

        if draft is None:
            return DOES_NOT_EXIST, {
                'reason': 'Review request draft does not exist.',
            }

        diffset = draft.diffset

        if diffset is None:
            return DOES_NOT_EXIST, {
                'reason': 'An empty diff must be created first.',
            }
        elif diffset.is_commit_series_finalized:
            return INVALID_ATTRIBUTE, {
                'reason': 'The diff has already been finalized.',
            }

        form = UploadCommitForm(
            review_request=review_request,
            diffset=diffset,
            request=request,
            data=request.POST.copy(),
            files=request.FILES)

        if not form.is_valid():
            return INVALID_FORM_DATA, {
                'fields': self._get_form_errors(form),
            }

        try:
            commit = form.create()
        except FileNotFoundError as e:
            return REPO_FILE_NOT_FOUND, {
                'file': e.path,
                'revision': str(e.revision),
            }
        except EmptyDiffError:
            return DIFF_EMPTY
        except DiffTooBigError as e:
            return DIFF_TOO_BIG, {
                'reason': str(e),
                'max_size': e.max_diff_size,
            }
        except Exception as e:
            logger.exception('Error uploading new commit: %s', e,
                             extra={'request': request})

            return INVALID_FORM_DATA, {
                'fields': {
                    'path': [str(e)],
                },
            }

        return 201, {
            self.item_result_key: commit,
        }

    def get_related_links(self, obj, request=None, *args, **kwargs):
        """Return the related links for the resource.

        If this is for an item resource, this will return links for all the
        associated FileDiffs.

        Args:
            obj (reviewboard.diffviewer.models.DiffCommit, optional):
                The DiffCommit to get links for.

            request (django.http.HttpRequest):
                The HTTP request from the client.

            *args (tuple):
                Additional positional arguments.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            dict:
            The related links.
        """
        links = {}

        if obj and request:
            links['draft_files'] = self._get_files_link(
                commit=obj,
                request=request,
                diff_resource=resources.draft_diff,
                filediff_resource=resources.draft_filediff,
                files_key='draft_files',
                *args,
                **kwargs)

        return links


draft_diffcommit_resource = DraftDiffCommitResource()



"""Base classes for actions.

Version Added:
    6.0
"""

from __future__ import annotations

import logging
from typing import Any, Dict, List, Optional, TYPE_CHECKING, cast

from django.template.loader import render_to_string
from django.utils.safestring import mark_safe
from djblets.pagestate.state import PageState
from housekeeping import func_deprecated
from typing_extensions import final

from reviewboard.actions.errors import (ActionError,
                                        MissingActionRendererError)
from reviewboard.actions.renderers import (BaseActionGroupRenderer,
                                           BaseActionRenderer,
                                           DefaultActionGroupRenderer,
                                           DefaultActionRenderer,
                                           MenuActionGroupRenderer)
from reviewboard.deprecation import RemovedInReviewBoard10_0Warning
from reviewboard.site.urlresolvers import local_site_reverse

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator, Sequence
    from typing import ClassVar

    from django.http import HttpRequest
    from django.template import Context
    from django.utils.safestring import SafeString
    from typelets.django.json import SerializableDjangoJSONDict
    from typelets.django.strings import StrOrPromise

    from reviewboard.actions.registry import ActionsRegistry


logger = logging.getLogger(__name__)


class AttachmentPoint:
    """Attachment points for actions.

    Version Added:
        6.0
    """

    #: Attachment for actions which do not want to render in the UI.
    #:
    #: This can be used for actions which want to be JavaScript-only, and may
    #: be used in the future for things like keyboard shortcuts or a command-K
    #: bar.
    NON_UI = 'non-ui'

    #: Administration navigation.
    #:
    #: Version Added:
    #:     8.0
    ADMIN_NAV = 'admin-nav'

    #: Attachment for actions in the page header.
    HEADER = 'header'

    #: Attachment for actions on the left side of the review request header.
    REVIEW_REQUEST_LEFT = 'review-request-left'

    #: Attachment for actions on the right side of the review request header.
    REVIEW_REQUEST = 'review-request'

    #: Attachment for actions in the unified draft banner.
    UNIFIED_BANNER = 'unified-banner'

    #: Attachment for actions in the quick access area.
    #:
    #: Version Added:
    #:     8.0
    QUICK_ACCESS = 'quick-access'


class ActionAttachmentPoint:
    """An attachment point for a list of actions.

    Attachment points manage the display and registration of the UI side of
    actions. They may contain any number of pre-defined actions, and will
    automatically contain any actions dynamically registered to the attachment
    point.

    Actions rendered in an attachment point will use their own default
    renderer if specified, and fall back to the attachment point's default
    renderer. This order ensures that an attachment point may, for instance,
    default to buttons but that a menu button can still be added and rendered
    correctly.

    An attachment point can be registered in a central registry so that it
    can be referred to by name. It may also be unregistered and passed directly
    to any templates that want to render the actions.

    To include an attachment point on a page, use the
    :py:func:`{% actions_html %}
    <reviewboard.actions.templatetags.actions.actions_html>` template tag.

    Version Added:
        8.0
    """

    ########################################
    # Instance-overridable class variables #
    ########################################

    #: A pre-defined list of action IDs to include on the attachment point.
    #:
    #: This may be set on the class or passed when initializing an instance.
    #:
    #: Any actions listed here will be listed first in the specified order.
    #: Any additional actions registered to this attachment point will be
    #: added after these actions.
    actions: (Sequence[str] | None) = None

    #: The unique ID for this attachment point.
    #:
    #: This may be set on the class or passed when initializing an instance.
    attachment_point_id: str

    #: The default action renderer used for child action groups.
    #:
    #: This may be set on the class or passed when initializing an instance.
    #:
    #: This will be used for any group actions that don't already have a
    #: default action renderer set. The action's default takes precedence in
    #: order to allow actions such as menus to manage their own display.
    default_action_group_renderer_cls: (
        type[BaseActionGroupRenderer] | None
    ) = DefaultActionGroupRenderer

    #: The default action renderer used for child actions.
    #:
    #: This may be set on the class or passed when initializing an instance.
    #:
    #: This will be used for any non-group actions that don't already have a
    #: default action renderer set. The action's default takes precedence
    #: in order to allow actions such as buttons to manage their own display.
    default_action_renderer_cls: (type[BaseActionRenderer] | None) = \
        DefaultActionRenderer

    ######################
    # Instance variables #
    ######################

    #: The registry managing actions for this attachment point.
    #:
    #: This is primarily for unit test purposes.
    actions_registry: ActionsRegistry

    def __init__(
        self,
        attachment_point_id: (str | None) = None,
        *,
        actions: (Sequence[str] | None) = None,
        actions_registry: (ActionsRegistry | None) = None,
        default_action_group_renderer_cls: (
            type[BaseActionGroupRenderer] | None
        ) = None,
        default_action_renderer_cls: (type[BaseActionRenderer] | None) = None,
    ) -> None:
        """Initialize the attachment point.

        Args:
            attachment_point_id (str, optional):
                The unique ID for this attachment point.

                This must be provided if :py:attr:`attachment_point_id` is
                not already set.

            actions (list of str, optional):
                An explicit list of action IDs to include in this attachment
                point.

                Any actions listed here will be listed first in the
                specified order. Any additional actions registered to this
                attachment point will be added after these actions.

            actions_registry (reviewboard.actions.registry.ActionsRegistry,
                              optional):
                The registry managing actions for this attachment point.

                This is primarily for unit test purposes.

            default_action_group_renderer_cls (type, optional):
                The default action renderer used for child action groups.

                This will be used for any actions that don't already have a
                default action renderer set. The action's default takes
                precedence in order to allow actions such as menus to manage
                their own display.

            default_action_renderer_cls (type, optional):
                The default action renderer used for child actions.

                This will be used for any actions that don't already have a
                default action renderer set. The action's default takes
                precedence in order to allow actions such as buttons to manage
                their own display.
        """
        if attachment_point_id:
            self.attachment_point_id = attachment_point_id
        elif not getattr(self, 'attachment_point_id', None):
            raise AttributeError(
                'attachment_point_id must be provided as an argument or a '
                'class attribute.'
            )

        if actions is not None:
            self.actions = actions

        if default_action_renderer_cls is not None:
            self.default_action_renderer_cls = default_action_renderer_cls

        if default_action_group_renderer_cls is not None:
            self.default_action_group_renderer_cls = \
                default_action_group_renderer_cls

        if not actions_registry:
            from reviewboard.actions import actions_registry

        assert actions_registry is not None
        self.actions_registry = actions_registry

    def iter_actions(
        self,
        *,
        include_children: bool = False,
    ) -> Iterator[BaseAction]:
        """Yield actions for this attachment point.

        Args:
            include_children (bool, optional):
                Whether to also include children of menus.

                If ``False`` (default), this will only yield the top-level
                items.

        Yields:
            reviewboard.actions.base.BaseAction:
            The actions for the given attachment point.
        """
        registered_actions = self.actions_registry.get_for_attachment(
            self.attachment_point_id,
            include_children=include_children,
        )

        # If there are any predefined order of actions, list those first.
        # Then list any registered actions after.
        if (actions := self.actions):
            actions_registry = self.actions_registry
            seen_actions: set[str] = set()

            for action_id in actions:
                action = actions_registry.get_action(action_id)

                if action is not None:
                    yield action
                else:
                    logger.warning(
                        'Attachment point "%s" referenced action "%s", but '
                        'that action was not registered.',
                        self.attachment_point_id, action_id)

                seen_actions.add(action_id)

            for action in registered_actions:
                if action.action_id not in seen_actions:
                    yield action
        else:
            # There are no predefined actions, so just return everything
            # registered at this attachment point.
            yield from registered_actions

    def get_js_view_data(
        self,
        *,
        context: Context,
    ) -> SerializableDjangoJSONDict:
        """Return additional data to be passed to the JavaScript view.

        This will be merged along with the data provided by the renderer.
        Any data in the renderer will take precedence over data returned
        by this method.

        Args:
            context (django.template.Context):
                The current rendering context.

        Returns:
            dict:
            A dictionary of attributes to pass to the model instance.
        """
        return {}

    def render(
        self,
        *,
        request: HttpRequest,
        context: Context,
    ) -> SafeString:
        """Render all actions in the attachment point.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            context (django.template.Context):
                The current rendering context.

        Returns:
            django.utils.safestring.SafeString:
            The rendered action HTML.
        """
        return mark_safe(''.join(self._iter_render(
            request=request,
            context=context,
        )))

    def render_js(
        self,
        *,
        request: HttpRequest,
        context: Context,
    ) -> SafeString:
        """Render the JavaScript for loading each action view in this point.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            context (django.template.Context):
                The current rendering context.

        Returns:
            django.utils.safestring.SafeString:
            The rendered action JavaScript.
        """
        return mark_safe(''.join(self._iter_render_js(
            request=request,
            context=context,
            actions=self.iter_actions(),
            default_action_renderer_cls=self.default_action_renderer_cls,
            default_action_group_renderer_cls=(
                self.default_action_group_renderer_cls),
        )))

    def _iter_render(
        self,
        *,
        request: HttpRequest,
        context: Context,
    ) -> Iterator[SafeString]:
        """Generate HTML for each top-level action in the attachment point.

        This will iterate through all the top-level actions and render each
        to HTML, yielding each one-by-one.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            context (django.template.Context):
                The current rendering context.

        Yields:
            django.utils.safestring.SafeString:
            Each rendered action HTML.
        """
        attachment_point_id = self.attachment_point_id
        default_renderer_classes = {
            True: self.default_action_group_renderer_cls,
            False: self.default_action_renderer_cls,
        }

        for action in self.iter_actions(include_children=False):
            placement = action.get_placement(attachment_point_id)
            renderer = action.get_renderer_cls(
                placement=placement,
                fallback_renderer_cls=(
                    default_renderer_classes[action._is_action_group]),
            )

            try:
                yield action.render(
                    request=request,
                    context=context,
                    placement=placement,
                    renderer=renderer,
                )
            except Exception as e:
                logger.exception('Error rendering action %r: %s',
                                 action, e,
                                 extra={'request': request})

    def _iter_render_js(
        self,
        *,
        request: HttpRequest,
        context: Context,
        actions: Iterable[BaseAction],
        default_action_renderer_cls: type[BaseActionRenderer] | None,
        default_action_group_renderer_cls: (type[BaseActionGroupRenderer] |
                                            None),
    ) -> Iterator[SafeString]:
        """Generate JavaScript for each action in the attachment point.

        This will iterate through all the actions (top-level and children)
        and render each action's JavaScript view registration, yielding each
        one-by-one.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            context (django.template.Context):
                The current rendering context.

            actions (list of BaseAction):
                The actions to iterate through.

            default_action_renderer_cls (type):
                The default action renderer to use as a fallback.

            default_action_group_renderer_cls (type):
                The default action group renderer to use as a fallback.

        Yields:
            django.utils.safestring.SafeString:
            Each rendered action JavaScript.
        """
        attachment_point_id = self.attachment_point_id
        js_view_data = self.get_js_view_data(context=context)
        default_renderer_classes = {
            True: default_action_group_renderer_cls,
            False: default_action_renderer_cls,
        }

        for action in actions:
            placement = action.get_placement(attachment_point_id)

            if not placement:
                continue

            renderer = action.get_renderer_cls(
                placement=placement,
                fallback_renderer_cls=(
                    default_renderer_classes[action._is_action_group]),
            )

            try:
                yield action.render_js(
                    request=request,
                    context=context,
                    placement=placement,
                    renderer=renderer,
                    extra_js_view_data=js_view_data,
                )
            except Exception as e:
                logger.exception('Error rendering JavaScript for '
                                 'action %r: %s',
                                 action, e,
                                 extra={'request': request})

                continue

            # Render any children using the item renderer as the default.
            if (action.should_render(context=context) and
                (child_actions := placement.child_actions)):
                if (renderer is not None and
                    issubclass(renderer, BaseActionGroupRenderer)):
                    child_default_item_renderer_cls = \
                        renderer.default_item_renderer_cls
                    child_default_subgroup_renderer_cls = \
                        renderer.default_subgroup_renderer_cls

                    # Check if this renderer wants to use itself to render
                    # subgroups.
                    if child_default_subgroup_renderer_cls == 'self':
                        child_default_subgroup_renderer_cls = renderer
                else:
                    # The current action's renderer is not a group renderer
                    # with defaults for child renderers. Because of this,
                    # we don't have a renderer we can use as a default. If
                    # any children have their own, they will be used, but
                    # any others may log an error.
                    child_default_item_renderer_cls = None
                    child_default_subgroup_renderer_cls = None

                yield from self._iter_render_js(
                    request=request,
                    context=context,
                    actions=child_actions,
                    default_action_renderer_cls=(
                        child_default_item_renderer_cls),
                    default_action_group_renderer_cls=(
                        child_default_subgroup_renderer_cls),
                )


@final
class ActionPlacement:
    """Placement information for an action.

    This is used to specify where and how an action may be placed. This is
    mapped to a key specifying an attachment point ID, and can specify the
    parent action within the attachment point and a default renderer to use.

    Version Added:
        8.0
    """

    ######################
    # Instance variables #
    ######################

    #: The attachment point for the action.
    attachment: str

    #: The list of child actions, if this is a grouped action.
    child_actions: list[BaseAction]

    #: The default renderer used for this action.
    #:
    #: By default, actions inherit their default renderer from a previous
    #: group or attachment point.
    #:
    #: Default renderers can always be overridden when rendering the action.
    default_renderer_cls: type[BaseActionRenderer] | None

    #: The DOM element ID for this element on the page.
    #:
    #: If provided, this must be unique to the action and attachment point
    #: to avoid ID collisions with other actions that are placed at the
    #: same attachment point.
    #:
    #: If not provided, an ID in the form of
    #: :samp:`action-{attachment}-{action_id}` will be used.
    dom_element_id: (str | None) = None

    #: The parent of this action, if this is an item in a group.
    parent_action: BaseGroupAction | None

    #: The ID of the parent action within the attachment point, if needed.
    #:
    #: This is used to build menus or groups of actions in part of the UI.
    parent_id: str | None

    def __init__(
        self,
        attachment: str,
        *,
        default_renderer_cls: (type[BaseActionRenderer] | None) = None,
        dom_element_id: (str | None) = None,
        parent_id: (str | None) = None,
    ) -> None:
        """Initialize the placement.

        Args:
            attachment (str):
                The attachment point ID to place the action in.

            default_renderer_cls (type, optional):
                The default renderer to use when rendering in this attachment
                point.

            dom_element_id (str, optional):
                A custom DOM element ID for the action in this attachment
                point.

                If provided, this must be unique to the action and attachment
                point to avoid ID collisions with other actions that are
                placed at the same attachment point.

            parent_id (str, optional):
                The parent ID of an action in this attachment point in which
                to place this action.
        """
        self.attachment = attachment
        self.default_renderer_cls = default_renderer_cls
        self.dom_element_id = dom_element_id
        self.parent_id = parent_id

        self.child_actions = []
        self.parent_action = None

    @property
    def depth(self) -> int:
        """The depth of the action in this placement.

        Type:
            int
        """
        if ((parent := self.parent_action) is None or
            (placement := parent.get_placement(self.attachment)) is None):
            return 0
        else:
            return placement.depth + 1

    def clone(self) -> ActionPlacement:
        """Return a clone of action placement.

        This is used to produce a copy of a central action that can be
        attached to an action instance, to avoid sharing of state.

        Returns:
            ActionPlacement:
            The cloned instance.
        """
        return ActionPlacement(
            attachment=self.attachment,
            default_renderer_cls=self.default_renderer_cls,
            dom_element_id=self.dom_element_id,
            parent_id=self.parent_id,
        )


class BaseAction:
    """Base class for actions.

    Version Changed:
        8.0:
        * A single action now can be placed in multiple attachment points
          via :py:attr:`placements`.

        * Actions now make use of renderer classes for rendering, instead
          of including rendering logic built-in.

        * Deprecated the attributes :py:attr:`attachment`,
          :py:attr:`js_view_class`, :py:attr:`parent_action`,
          :py:attr:`parent_id`, and :py:attr:`template_name`.

        * Deprecated the methods :py:meth:`get_js_view_data` and
          :py:meth:`is_custom_rendered`.

    Version Added:
        6.0
    """

    #: The internal ID of the action.
    #:
    #: This must be unique.
    #:
    #: Type:
    #:     str
    action_id: str

    #: A list of URLs to apply to.
    #:
    #: If this is ``None``, the action will be loaded on all pages. Otherwise,
    #: it will be limited to the URLs listed here.
    #:
    #: Type:
    #:     list of str
    apply_to: Optional[List[str]] = None

    #: The attachment point for the action.
    #:
    #: Deprecated:
    #:     8.0:
    #:     This has been replaced by :py:attr:`placements`. This option
    #:     will be removed in Review Board 9.
    attachment: (str | None) = None

    #: The default renderer used for this action.
    #:
    #: By default, actions inherit their default renderer from a previous
    #: group or attachment point.
    #:
    #: Default renderers can always be overridden when rendering the action.
    #:
    #: Version Added:
    #:     8.0
    default_renderer_cls: (type[BaseActionRenderer] | None) = None

    #: An optional description for your action.
    #:
    #: If provided, this must be a list of strings, each used as a paragraph.
    #: Each must be plain text, and will be escaped for display.
    #:
    #: When rendered as a button, this will be used for the button's tooltip.
    #:
    #: When rendered as a detailed menu item (which is used in the
    #: :ref:`Review menu <review-banner-menu>`), this will be shown as
    #: additional text below the label.
    #:
    #: Other renderers may use this for other purposes.
    #:
    #: Version Added:
    #:     8.0
    description: (Sequence[StrOrPromise] | None) = None

    #: A class name to use for an icon.
    #:
    #: If specified, this should be the entire class to apply to a <span>
    #: element to display an icon. For example, 'fa fa-rss'.
    #:
    #: Type:
    #:     str
    icon_class: Optional[str] = None

    #: The class to instantiate for the JavaScript model.
    #:
    #: Type:
    #:     str
    js_model_class: str = 'RB.Actions.Action'

    #: The name of the template to use for registering action model JavaScript.
    js_template_name: str = 'actions/action.js'

    #: The class to instantiate for the JavaScript view.
    #:
    #: Deprecated:
    #:     8.0:
    #:     This should be set on the action's renderer instead.
    js_view_class: (str | None) = None

    #: The user-visible label.
    #:
    #: Type:
    #:     str
    label: (StrOrPromise | None) = None

    #: The IDs of any parent actions, if needed.
    #:
    #: Deprecated:
    #:     8.0:
    #:     This should be set in :py:attr:`placements` instead. This option
    #:     will be removed in Review Board 9.
    parent_id: (str | None) = None

    #: The placements of this action within the page.
    #:
    #: Each entry defines a placement within an attachment point and an
    #: action parent/child hierarchy where action should be rendered, along
    #: with options controlling the rendering.
    #:
    #: Version Added:
    #:     8.0
    placements: (Sequence[ActionPlacement] | None) = None

    #: The name of the template to use when rendering.
    #:
    #: Deprecated:
    #:     8.0:
    #:     This should be set on the action's renderer instead. This option
    #:     will be removed in Review Board 9.
    template_name: (str | None) = None

    #: The URL that this action links to.
    #:
    #: Type:
    #:     str
    url: str = '#'

    #: A URL name to resolve.
    #:
    #: If this is not None, it will take precedence over :py:attr:`url`.
    #:
    #: Type:
    #:     str
    url_name: Optional[str] = None

    #: The user-visible verbose label.
    #:
    #: This can be used to provide a longer label for wider UIs that would
    #: benefit from a more descriptive label. It's also intended for ARIA
    #: labels.
    #:
    #: This is always optional.
    #:
    #: Version Added:
    #:     8.0
    verbose_label: (StrOrPromise | None) = None

    #: Whether this action is visible.
    #:
    #: Type:
    #:     bool
    visible: bool = True

    #: Whether or not this represents a group of actions.
    #:
    #: This allows for an inexpensive way of checking the base type of an
    #: action without either performing instanceof/subclass checks. It should
    #: be considered an internal flag and not subject to API compatibility
    #: guarantees.
    #:
    #: Version Added:
    #:     8.0
    _is_action_group: ClassVar[bool] = False

    ######################
    # Instance variables #
    ######################

    #: The parent registry managing this action.
    #:
    #: Version Added:
    #:     8.0
    parent_registry: ActionsRegistry | None

    #: A cache of attachment point IDs to placement information.
    #:
    #: Version Added:
    #:     8.0
    _placement_cache: dict[str, ActionPlacement] | None

    def __init__(self) -> None:
        """Initialize the action."""
        cls = type(self)

        self.parent_registry = None
        self._placement_cache = None

        if not getattr(self, 'action_id', None):
            raise AttributeError(
                f'{cls.__name__}.action_id must be set.'
            )

        attachment = self.attachment
        parent_id = self.parent_id

        if hasattr(self, '_ignore_action_deprecations'):
            # This is a compatibility subclass. We don't want to check
            # the deprecations below.
            if not self.placements:
                self.placements = [
                    ActionPlacement(
                        attachment=(attachment or
                                    AttachmentPoint.REVIEW_REQUEST),
                        parent_id=parent_id),
                ]

            return

        # Check for deprecations.
        if attachment:
            RemovedInReviewBoard10_0Warning.warn(
                f'{type(self).__name__}.attachment is deprecated and '
                f'support will be removed in Review Board 10. Please set the '
                f'"placements" attribute instead.'
            )

            self.placements = [
                ActionPlacement(attachment=attachment,
                                parent_id=parent_id),
            ]
        elif not self.placements:
            self.placements = [
                ActionPlacement(attachment=AttachmentPoint.REVIEW_REQUEST,
                                parent_id=parent_id),
            ]
        else:
            # Copy all the placements, so we don't share state between
            # multiple actions or subclasses of actions that define the
            # same placements.
            self.placements = [
                placement.clone()
                for placement in self.placements
            ]

        if parent_id:
            RemovedInReviewBoard10_0Warning.warn(
                f'{type(self).__name__}.parent_id is deprecated and '
                f'support will be removed in Review Board 10. Please set this '
                f'attribute in "placements" instead.'
            )

        if self.template_name:
            RemovedInReviewBoard10_0Warning.warn(
                f'{cls.__name__}.template_name is deprecated and '
                f'support will be removed in Review Board 10. Please move any '
                f'custom rendering to a reviewboard.actions.renderers.'
                f'BaseActionRenderer subclass instead.'
            )

        if self.js_view_class:
            RemovedInReviewBoard10_0Warning.warn(
                f'{cls.__name__}.js_view_class is deprecated and '
                f'support will be removed in Review Board 10. Please move any '
                f'custom rendering to a reviewboard.actions.renderers.'
                f'BaseActionRenderer subclass instead.'
            )

        if cls.get_js_view_data is not BaseAction.get_js_view_data:
            RemovedInReviewBoard10_0Warning.warn(
                f'{cls.__name__}.get_js_view_data is deprecated and '
                f'support will be removed in Review Board 10. Please move any '
                f'custom rendering to a reviewboard.actions.renderers.'
                f'BaseActionRenderer subclass instead.'
            )

        if cls.get_dom_element_id is not BaseAction.get_dom_element_id:
            RemovedInReviewBoard10_0Warning.warn(
                f'{cls.__name__}.get_dom_element_id is deprecated and '
                f'support will be removed in Review Board 10.'
            )

    @property
    @func_deprecated(RemovedInReviewBoard10_0Warning, message=(
        '%(func_name)s is deprecated and support will be removed in '
        '%(product)s %(version)s. Please use ActionPlacement.parent_action '
        'instead.'
    ))
    def parent_action(self) -> BaseAction | None:
        """The parent of this action, if this is an item in a group.

        Deprecated:
            8.0:
            This has been replaced by
            :py:attr:`ActionPlacement.parent_action`. It will be removed in
            Review Board 10.
        """
        if (placements := self.placements):
            return placements[0].parent_action

        return None

    @property
    @func_deprecated(RemovedInReviewBoard10_0Warning, message=(
        '%(func_name)s is deprecated and support will be removed in '
        '%(product)s %(version)s. Please use ActionPlacement.child_actions '
        'instead.'
    ))
    def child_actions(self) -> Sequence[BaseAction]:
        """The children of this action, if this is a group.

        Deprecated:
            8.0:
            This has been replaced by
            :py:attr:`ActionPlacement.child_actions`. It will be removed in
            Review Board 10.
        """
        if (placements := self.placements):
            return placements[0].child_actions

        return []

    @property
    @func_deprecated(RemovedInReviewBoard10_0Warning, message=(
        '%(func_name)s is deprecated and support will be removed in '
        '%(product)s %(version)s. Please use ActionPlacement.depth '
        'instead.'
    ))
    def depth(self) -> int:
        """The depth of the action.

        Deprecated:
            8.0:
            This is scheduled for removal in Review Board 10. This has been
            replaced by :py:attr:`ActionPlacement.depth`.

        Type:
            int
        """
        if (placements := self.placements):
            return placements[0].depth

        return 0

    def get_placement(
        self,
        attachment_point_id: str,
    ) -> ActionPlacement:
        """Return the Placement for the action matching the attachment point.

        Version Added:
            8.0

        Args:
            attachment_point_id (str):
                The ID of the attachment point matching the placement.

        Returns:
            ActionPlacement:
            The Placement, or ``None`` if not found.

        Raises:
            KeyError:
                The attachment point ID is not a valid placement for this
                action.
        """
        placement_cache = self._placement_cache

        if placement_cache is None:
            placement_cache = {
                placement.attachment: placement
                for placement in self.placements or []
            }
            self._placement_cache = placement_cache

        try:
            return placement_cache[attachment_point_id]
        except KeyError:
            raise KeyError(
                f'"{attachment_point_id}" is not a valid placement for '
                f'action "{self.action_id}".'
            )

    def get_renderer_cls(
        self,
        *,
        placement: ActionPlacement | None,
        preferred_renderer_cls: (type[BaseActionRenderer] | None) = None,
        fallback_renderer_cls: (type[BaseActionRenderer] | None) = None,
    ) -> type[BaseActionRenderer] | None:
        """Return the renderer class used to render this action.

        This takes into account the attachment point, the preferred
        renderer from the caller, and the fallback if no suitable renderer
        is found.

        Version Added:
            8.0

        Args:
            placement (ActionPlacement):
                The placement where the action will be rendered.

            preferred_renderer_cls (type, optional):
                The preferred renderer for items.

                This will be used if provided, unless item isn't
                custom-rendered (a deprecated feature).

            fallback_renderer_cls (type, optional):
                The fallback renderer if the placement or action does not
                provide one.

        Returns:
            type:
            The renderer class, or ``None`` if one could not be determined.
        """
        if preferred_renderer_cls and not self.is_custom_rendered():
            return preferred_renderer_cls

        if placement is not None and placement.default_renderer_cls:
            return placement.default_renderer_cls

        return self.default_renderer_cls or fallback_renderer_cls

    def is_custom_rendered(self) -> bool:
        """Whether this action uses custom rendering.

        By default, this will return ``True`` if a custom template name is
        used. If the JavaScript side needs to override rendering, the subclass
        should explicitly return ``True``.

        Deprecated:
            8.0:
            This is scheduled for removal in Review Board 9. This was only
            ever used for menu items. Custom menu items should instead set
            the ``data-custom-rendered="true"`` attribute on the custom
            element.

        Version Added:
            7.0

        Returns:
            bool:
            ``True`` if this action uses custom rendering. ``False`` if it
            does not.
        """
        return self.template_name != BaseAction.template_name

    def should_register(
        self,
        *,
        context: Context,
        request: HttpRequest,
    ) -> bool:
        """Return whether this action should be registered.

        By default, an action will be registered only if it's rendered
        somewhere on the page (which will itself depend in part on the
        result of :py:meth:`should_render`).

        Subclasses can override this to register even if not rendered, which
        may be useful for actions that are entirely managed dynamically in
        JavaScript. When overriding, it's recommended to call
        :py:meth:`should_render` in order to avoid registering an action
        that is not applicable to a page or is hidden by a hook.

        Version Added:
            8.0

        Args:
            context (django.template.Context):
                The current rendering context.

            request (django.http.HttpRequest):
                The HTTP request from the client.

        Returns:
            bool:
            ``True`` if the action should be registered. ``False`` if it
            should not.
        """
        page_state = PageState.for_request(request)

        try:
            rendered_action_ids = page_state.extra_data['rb-actions-rendered']
        except KeyError:
            # This should only happen during tests, but safeguard anyway.
            return False

        return self.action_id in rendered_action_ids

    def should_render(
        self,
        *,
        context: Context,
    ) -> bool:
        """Return whether this action should render.

        This differs from :py:attr:`visible` in that non-visible actions still
        render but are hidden by CSS, whereas if this returns ``False`` the
        action will not be included in the DOM at all.

        Args:
            context (django.template.Context):
                The current rendering context.

        Returns:
            bool:
            ``True`` if the action should render.
        """
        request = context['request']

        if ((apply_to := self.apply_to) and not
            (request.resolver_match and
             request.resolver_match.url_name in apply_to)):
            return False

        from reviewboard.extensions.hooks.actions import HideActionHook

        action_id = self.action_id

        for hook in HideActionHook.hooks:
            if action_id in hook.hidden_action_ids:
                return False

        return True

    @func_deprecated(RemovedInReviewBoard10_0Warning, message=(
        '%(func_name)s is deprecated and support will be removed in '
        '%(product)s %(version)s. Please use the dom_element_id context '
        'variable in templates or set ActionPlacement.dom_element_id instead.'
    ))
    def get_dom_element_id(self) -> str:
        """Return the ID used for the DOM element for this action.

        Deprecated:
            8.0:
            This is scheduled for removal in Review Board 10. This has been
            replaced by :py:attr:`ActionPlacement.dom_element_id`.

        Returns:
            str:
            The ID used for the element.
        """
        # We still need to return a value here instead of an empty string.
        # Otherwise we break legacy extensions that have templates calling
        # action.get_dom_element_id(), which is a likely scenario because a
        # lot of our action templates did this.
        if placements := self.placements:
            # Match the ID that the renderer will use. Legacy actions are
            # single-placement, so it's safe to use the first placement.
            placement = placements[0]

            if placement.dom_element_id:
                return placement.dom_element_id

            attachment = placement.attachment
        else:
            # Actions always have at least one placement once they've been
            # initialized, so this is effectively unreachable. Fall back to
            # the default attachment point though just in case.
            attachment = AttachmentPoint.REVIEW_REQUEST

        return f'action-{attachment}-{self.action_id}'

    def get_js_model_data(
        self,
        *,
        context: Context,
    ) -> SerializableDjangoJSONDict:
        """Return data to be passed to the JavaScript model.

        Args:
            context (django.template.Context):
                The current rendering context.

        Returns:
            dict:
            A dictionary of attributes to pass to the model instance.
        """
        icon_class = self.icon_class

        try:
            url = self.get_url(context=context)
        except Exception as e:
            logger.exception('Unexpected error retrieving URL for action '
                             '"%s": %s',
                             self.action_id, e)
            url = None

        visible = self.get_visible(context=context)

        data: SerializableDjangoJSONDict = {
            'id': self.action_id,
            'visible': visible,
        }

        if icon_class:
            data['iconClass'] = icon_class

        if self.is_custom_rendered():
            data['isCustomRendered'] = True

        if (label := self.get_label(context=context)):
            data['label'] = str(label)

        if (verbose_label := self.get_verbose_label(context=context)):
            data['verboseLabel'] = str(verbose_label)

        if url:
            data['url'] = url

        return data

    def get_js_view_data(
        self,
        *,
        context: Context,
    ) -> SerializableDjangoJSONDict:
        """Return data to be passed to the JavaScript view.

        Deprecated:
            8.0:
            Actions implementing this should instead move to custom
            renderers.

        Args:
            context (django.template.Context):
                The current rendering context.

        Returns:
            dict:
            A dictionary of options to pass to the view instance.
        """
        return {}

    def get_label(
        self,
        *,
        context: Context,
    ) -> StrOrPromise | None:
        """Return the label for the action.

        Args:
            context (django.template.Context):
                The current rendering context.

        Returns:
            str:
            The label to use for the action.
        """
        return self.label

    def get_verbose_label(
        self,
        *,
        context: Context,
    ) -> StrOrPromise | None:
        """Return the verbose label for the action.

         This can be used to provide a longer label for wider UIs that would
         benefit from a more descriptive label. It's always optional.

         Version Added:
             8.0

        Args:
            context (django.template.Context):
                The current rendering context.

        Returns:
            str:
            The verbose label to use for the action.
        """
        return self.verbose_label

    def get_url(
        self,
        *,
        context: Context,
    ) -> str:
        """Return the URL for the action.

        Args:
            context (django.template.Context):
                The current rendering context.

        Returns:
            str:
            The URL to use for the action.
        """
        assert self.url_name or self.url

        if self.url_name:
            return local_site_reverse(self.url_name,
                                      request=context.get('request'))
        else:
            return self.url

    def get_visible(
        self,
        *,
        context: Context,
    ) -> bool:
        """Return whether the action should start visible.

        Args:
            context (django.template.Context):
                The current rendering context.

        Returns:
            bool:
            ``True`` if the action should start visible. ``False``, otherwise.
        """
        return self.visible

    def get_extra_context(
        self,
        *,
        request: HttpRequest,
        context: Context,
    ) -> dict[str, Any]:
        """Return extra template context for the action.

        Subclasses can override this to provide additional context needed by
        the template for the action. By default, this returns an empty
        dictionary.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            context (django.template.Context):
                The current rendering context.

        Returns:
            dict:
            Extra context to use when rendering the action's template.
        """
        return {
            'id': self.action_id,
            'label': self.get_label(context=context),
            'url': self.get_url(context=context),
            'verbose_label': self.get_verbose_label(context=context),
            'visible': self.get_visible(context=context),
        }

    def render(
        self,
        *,
        request: HttpRequest,
        context: Context,
        placement: (ActionPlacement | None) = None,
        renderer: (type[BaseActionRenderer] | None) = None,
        fallback_renderer: (type[BaseActionRenderer] | None) = None,
    ) -> SafeString:
        """Render the action.

        Version Changed:
            8.0:
            Added the ``extra_js_view_data``, ``fallback_renderer``,
            ``placement``, and ``renderer`` arguments.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            context (django.template.Context):
                The current rendering context.

            placement (ActionPlacement, optional):
                The placement the action is being rendered into.

                If not added, the first will be used, with a warning.

                This argument will be required in Review Boad 9.

                Version Added:
                    8.0

            renderer (type, optional):
                The renderer used to render this action.

                If not specified, :py:attr:`default_renderer_cls` will be
                used.

                Version Added:
                    8.0

            fallback_renderer (type, optional):
                The renderer used to render this action if no other is found.

                Version Added:
                    8.0

        Returns:
            django.utils.safestring.SafeString:
            The rendered action HTML.

        Raises:
            TypeError:
                An invalid renderer class was provided.

            reviewboard.actions.errors.MissingActionRendererError:
                A renderer was not found or provided.
        """
        if not self.should_render(context=context):
            return mark_safe('')

        if placement is None:
            RemovedInReviewBoard10_0Warning.warn(
                f'{type(self).__name__}.render() must be passed a placement= '
                f'argument. This will be required in Review Board 10.'
            )

            if not (placements := self.placements):
                return mark_safe('')

            placement = placements[0]

        renderer_cls = self.get_renderer_cls(
            placement=placement,
            preferred_renderer_cls=renderer,
            fallback_renderer_cls=fallback_renderer,
        )

        if renderer_cls is None:
            raise MissingActionRendererError(
                f'A renderer must be explicitly provided when rendering '
                f'action {type(self)!r}.'
            )

        if not issubclass(renderer_cls, BaseActionRenderer):
            raise TypeError(
                f'An invalid renderer class was provided ({renderer_cls!r}).'
            )

        if not self.should_render(context=context):
            return mark_safe('')

        return (
            renderer_cls(action=self,
                         placement=placement)
            .render(request=request,
                    context=context)
        )

    def render_model_js(
        self,
        *,
        request: HttpRequest,
        context: Context,
    ) -> SafeString:
        """Render the action's JavaScript model.

        This will set up the JavaScript model, registering it for use in
        views.

        Version Added:
            8.0

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            context (django.template.Context):
                The current rendering context.

        Returns:
            django.utils.safestring.SafeString:
            The rendered action JavaScript.
        """
        with context.push():
            try:
                context.update({
                    'action': self,
                    'js_model_class': self.js_model_class,
                })

                return render_to_string(
                    template_name=self.js_template_name,
                    context=cast(Dict[str, Any], context.flatten()),
                    request=request)
            except Exception as e:
                logger.exception('Error rendering JavaScript for action model '
                                 '%r: %s',
                                 self, e)

                return mark_safe('')

    def render_js(
        self,
        *,
        request: HttpRequest,
        context: Context,
        extra_js_view_data: (SerializableDjangoJSONDict | None) = None,
        placement: (ActionPlacement | None) = None,
        renderer: (type[BaseActionRenderer] | None) = None,
        fallback_renderer: (type[BaseActionRenderer] | None) = None,
    ) -> SafeString:
        """Render the action's JavaScript view.

        This will set up an instance of an action's view using either the
        provided renderer or the action's default renderer.

        Version Changed:
            8.0:
            Added the ``extra_js_view_data``, ``fallback_renderer``,
            ``placement``, and ``renderer`` arguments.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            context (django.template.Context):
                The current rendering context.

            extra_js_view_data (dict, optional):
                Optional extra data to pass to the JavaScript action view's
                constructor.

                Version Added:
                    8.0

            placement (ActionPlacement, optional):
                The placement the action is being rendered into.

                If not added, the first will be used, with a warning.

                This argument will be required in Review Boad 9.

                Version Added:
                    8.0

            renderer (type, optional):
                The renderer used to render this action.

                If not specified, :py:attr:`default_renderer_cls` will be
                used.

                Version Added:
                    8.0

            fallback_renderer (type, optional):
                The renderer used to render this action if no other is found.

                Version Added:
                    8.0

        Returns:
            django.utils.safestring.SafeString:
            The rendered action JavaScript.

        Raises:
            TypeError:
                An invalid renderer class was provided.

            reviewboard.actions.errors.MissingActionRendererError:
                A renderer was not found or provided.
        """
        if not self.should_render(context=context):
            return mark_safe('')

        if placement is None:
            RemovedInReviewBoard10_0Warning.warn(
                f'{type(self).__name__}.render_js() must be passed a '
                f'placement= argument. This will be required in '
                f'Review Board 10.'
            )

            if not (placements := self.placements):
                return mark_safe('')

            placement = placements[0]

        renderer_cls = self.get_renderer_cls(
            placement=placement,
            preferred_renderer_cls=renderer,
            fallback_renderer_cls=fallback_renderer,
        )

        if renderer_cls is None:
            raise MissingActionRendererError(
                f'A renderer must be explicitly provided when rendering '
                f'JavaScript for action {type(self)!r}.'
            )

        if not issubclass(renderer_cls, BaseActionRenderer):
            raise TypeError(
                f'An invalid renderer class was provided ({renderer_cls!r}).'
            )

        return (
            renderer_cls(action=self,
                         placement=placement)
            .render_js(request=request,
                       context=context,
                       extra_js_view_data=extra_js_view_data)
        )


class BaseGroupAction(BaseAction):
    """Base class for a group of actions.

    This can be used to group together actions in some form. Subclasses
    can implement this as menus, lists of actions, or in other
    presentational styles.

    Version Added:
        8.0
    """

    _is_action_group = True
    js_model_class = 'RB.Actions.GroupAction'

    #: An ordered list of child action IDs.
    #:
    #: This can be used to specify a specific order for children to appear in.
    #: The special string '--' can be used to add separators. Any children that
    #: are registered with this group as their parent but do not appear in this
    #: list will be added at the end of the group.
    children: Sequence[str] = []

    def get_js_model_data(
        self,
        *,
        context: Context,
    ) -> SerializableDjangoJSONDict:
        """Return data to be passed to the JavaScript model.

        Args:
            context (django.template.Context):
                The current rendering context.

        Returns:
            dict:
            A dictionary of attributes to pass to the model instance.

        Raises:
            reviewboard.actions.errors.ActionError:
                There was an error retrieving data for the action.

                Details will be in the error message.
        """
        registry = self.parent_registry

        if not registry:
            raise ActionError(
                f'Attempted to call get_js_model_data on {self!r} without '
                f'first being registered.'
            )

        action_id = self.action_id
        assert action_id is not None

        action_children = self.children
        placement_children: dict[str, list[str]] = {}

        for placement in (self.placements or []):
            # Note that we're using a dict here instead of a set in order to
            # maintain order.
            placement_child_ids: dict[str, bool] = {
                child.action_id: True
                for child in placement.child_actions
                if child.action_id and child.should_render(context=context)
            }

            children: list[str] = []

            # Add in any children with explicit ordering first.
            for child_id in action_children:
                if child_id == '--':
                    children.append(child_id)
                elif child_id in placement_child_ids:
                    children.append(child_id)
                    del placement_child_ids[child_id]

            # Now add any other actions that weren't in self.children.
            children += placement_child_ids.keys()

            placement_children[placement.attachment] = children

        data = super().get_js_model_data(context=context)
        data['children'] = placement_children

        return data


class BaseMenuAction(BaseGroupAction):
    """Base class for menu actions.

    Version Added:
        6.0
    """

    default_renderer_cls: (type[BaseActionRenderer] | None) = \
        MenuActionGroupRenderer
    js_model_class = 'RB.Actions.MenuAction'

    def is_custom_rendered(self) -> bool:
        """Whether this menu action uses custom rendering.

        By default, this will return ``True`` if a custom template name is
        used. If the JavaScript side needs to override rendering, the subclass
        should explicitly return ``True``.

        Deprecated:
            8.0:
            This is scheduled for removal in Review Board 10. This was only
            ever used for menu items. Custom menu items should instead set
            the ``data-custom-rendered="true"`` attribute on the custom
            element.

        Version Added:
            7.0

        Returns:
            bool:
            ``True`` if this action uses custom rendering. ``False`` if it
            does not.
        """
        return self.template_name != BaseMenuAction.template_name



"""Unit tests for reviewboard.accounts.templatetags.accounts."""

from datetime import datetime

import pytz
from django.contrib.auth.models import AnonymousUser, User
from django.template import Context, Template
from django.test.client import RequestFactory
from django.utils.safestring import SafeText
from djblets.testing.decorators import add_fixtures

from reviewboard.accounts.user_details import (BaseUserDetailsProvider,
                                               UserBadge,
                                               user_details_provider_registry)
from reviewboard.testing import TestCase


class JSUserSessionInfoTests(TestCase):
    """Unit tests for {% js_user_session_info %}."""

    maxDiff = 100000

    def setUp(self):
        super(JSUserSessionInfoTests, self).setUp()

        self.user = self.create_user(username='test',
                                     first_name='Test',
                                     last_name='User',
                                     email='test@example.com')

    def test_with_anonymous_user(self):
        """Testing {% js_user_session_info %} with anonymous user"""
        self.assertJSONEqual(
            self._render_tag(AnonymousUser()),
            {
                'authenticated': False,
                'loginURL': '/account/login/',
            })

    def test_with_authenticated_user(self):
        """Testing {% js_user_session_info %} with authenticated user"""
        profile = self.user.get_profile()
        profile.timezone = 'US/Pacific'
        profile.save(update_fields=('timezone',))

        tz = pytz.timezone('US/Pacific')

        if tz.dst(datetime.now()):
            expected_tz_offset = '-0700'
        else:
            expected_tz_offset = '-0800'

        avatar_url = ('https://secure.gravatar.com/avatar/'
                      '55502f40dc8b7c769880b10874abc9d0')

        self.assertJSONEqual(
            self._render_tag(self.user),
            {
                'allowSelfShipIt': True,
                'archivedReviewRequestsURL':
                    '/api/users/test/archived-review-requests/',
                'authenticated': True,
                'avatarHTML': {
                    '32': '<img src="https://secure.gravatar.com/avatar/'
                          '55502f40dc8b7c769880b10874abc9d0?s=32&d=mm"'
                          ' alt="Test User" width="32" height="32"'
                          ' srcset="https://secure.gravatar.com/avatar/'
                          '55502f40dc8b7c769880b10874abc9d0?s=32&d=mm 1x,'
                          ' https://secure.gravatar.com/avatar/'
                          '55502f40dc8b7c769880b10874abc9d0?s=64&d=mm 2x,'
                          ' https://secure.gravatar.com/avatar/'
                          '55502f40dc8b7c769880b10874abc9d0?s=96&d=mm 3x"'
                          ' class="avatar djblets-o-avatar">\n',
                },
                'avatarURLs': {
                    '32': {
                        '1x': '%s?s=32&d=mm' % avatar_url,
                        '2x': '%s?s=64&d=mm' % avatar_url,
                        '3x': '%s?s=96&d=mm' % avatar_url,
                    },
                },
                'commentsOpenAnIssue': True,
                'confirmShipIt': True,
                'defaultUseRichText': True,
                'enableDesktopNotifications': True,
                'enableSpellChecking': None,
                'fullName': 'Test User',
                'mutedReviewRequestsURL':
                    '/api/users/test/muted-review-requests/',
                'quickAccessActionIDs': [],
                'readOnly': False,
                'sessionURL': '/api/session/',
                'timezoneOffset': expected_tz_offset,
                'userFileAttachmentsURL':
                    '/api/users/test/user-file-attachments/',
                'userPageURL': '/users/test/',
                'username': 'test',
                'watchedReviewGroupsURL':
                    '/api/users/test/watched/review-groups/',
                'watchedReviewRequestsURL':
                    '/api/users/test/watched/review-requests/',
            })

    @add_fixtures(['test_site', 'test_users'])
    def test_with_authenticated_user_and_local_site(self):
        """Testing {% js_user_session_info %} with authenticated user and
        LocalSite
        """
        profile = self.user.get_profile()
        profile.timezone = 'US/Pacific'
        profile.save(update_fields=('timezone',))

        tz = pytz.timezone('US/Pacific')

        if tz.dst(datetime.now()):
            expected_tz_offset = '-0700'
        else:
            expected_tz_offset = '-0800'

        local_site = self.get_local_site('local-site-1')

        avatar_url = ('https://secure.gravatar.com/avatar/'
                      '55502f40dc8b7c769880b10874abc9d0')

        self.assertJSONEqual(
            self._render_tag(self.user, local_site=local_site),
            {
                'allowSelfShipIt': True,
                'archivedReviewRequestsURL':
                    '/s/local-site-1/api/users/test/archived-review-requests/',
                'authenticated': True,
                'avatarHTML': {
                    '32': '<img src="https://secure.gravatar.com/avatar/'
                          '55502f40dc8b7c769880b10874abc9d0?s=32&d=mm"'
                          ' alt="Test User" width="32" height="32"'
                          ' srcset="https://secure.gravatar.com/avatar/'
                          '55502f40dc8b7c769880b10874abc9d0?s=32&d=mm 1x,'
                          ' https://secure.gravatar.com/avatar/'
                          '55502f40dc8b7c769880b10874abc9d0?s=64&d=mm 2x,'
                          ' https://secure.gravatar.com/avatar/'
                          '55502f40dc8b7c769880b10874abc9d0?s=96&d=mm 3x"'
                          ' class="avatar djblets-o-avatar">\n',
                },
                'avatarURLs': {
                    '32': {
                        '1x': '%s?s=32&d=mm' % avatar_url,
                        '2x': '%s?s=64&d=mm' % avatar_url,
                        '3x': '%s?s=96&d=mm' % avatar_url,
                    },
                },
                'commentsOpenAnIssue': True,
                'confirmShipIt': True,
                'defaultUseRichText': True,
                'enableDesktopNotifications': True,
                'enableSpellChecking': None,
                'fullName': 'Test User',
                'mutedReviewRequestsURL':
                    '/s/local-site-1/api/users/test/muted-review-requests/',
                'quickAccessActionIDs': [],
                'readOnly': False,
                'sessionURL': '/s/local-site-1/api/session/',
                'timezoneOffset': expected_tz_offset,
                'userFileAttachmentsURL':
                    '/s/local-site-1/api/users/test/user-file-attachments/',
                'userPageURL': '/s/local-site-1/users/test/',
                'username': 'test',
                'watchedReviewGroupsURL':
                    '/s/local-site-1/api/users/test/watched/review-groups/',
                'watchedReviewRequestsURL':
                    '/s/local-site-1/api/users/test/watched/review-requests/',
            })

    def test_with_authenticated_and_no_profile(self):
        """Testing {% js_user_session_info %} with authenticated user and
        no Profile
        """
        self.user.get_profile().delete()

        avatar_url = ('https://secure.gravatar.com/avatar/'
                      '55502f40dc8b7c769880b10874abc9d0')

        siteconfig_settings = {
            'default_use_rich_text': False,
        }

        with self.siteconfig_settings(siteconfig_settings):
            self.assertJSONEqual(
                self._render_tag(self.user),
                {
                    'allowSelfShipIt': True,
                    'archivedReviewRequestsURL':
                        '/api/users/test/archived-review-requests/',
                    'authenticated': True,
                    'avatarHTML': {
                        '32': '<img src="https://secure.gravatar.com/avatar/'
                              '55502f40dc8b7c769880b10874abc9d0?s=32&d=mm"'
                              ' alt="Test User" width="32" height="32"'
                              ' srcset="https://secure.gravatar.com/avatar/'
                              '55502f40dc8b7c769880b10874abc9d0?s=32&d=mm 1x,'
                              ' https://secure.gravatar.com/avatar/'
                              '55502f40dc8b7c769880b10874abc9d0?s=64&d=mm 2x,'
                              ' https://secure.gravatar.com/avatar/'
                              '55502f40dc8b7c769880b10874abc9d0?s=96&d=mm 3x"'
                              ' class="avatar djblets-o-avatar">\n',
                    },
                    'avatarURLs': {
                        '32': {
                            '1x': '%s?s=32&d=mm' % avatar_url,
                            '2x': '%s?s=64&d=mm' % avatar_url,
                            '3x': '%s?s=96&d=mm' % avatar_url,
                        },
                    },
                    'commentsOpenAnIssue': True,
                    'confirmShipIt': True,
                    'defaultUseRichText': False,
                    'enableDesktopNotifications': True,
                    'enableSpellChecking': None,
                    'fullName': 'Test User',
                    'mutedReviewRequestsURL':
                        '/api/users/test/muted-review-requests/',
                    'quickAccessActionIDs': [],
                    'readOnly': False,
                    'sessionURL': '/api/session/',
                    'timezoneOffset': '+0000',
                    'userFileAttachmentsURL':
                        '/api/users/test/user-file-attachments/',
                    'userPageURL': '/users/test/',
                    'username': 'test',
                    'watchedReviewGroupsURL':
                        '/api/users/test/watched/review-groups/',
                    'watchedReviewRequestsURL':
                        '/api/users/test/watched/review-requests/',
                })

    def test_with_authenticated_and_avatars_disabled(self):
        """Testing {% js_user_session_info %} with authenticated user and
        avatars disabled
        """
        siteconfig_settings = {
            'avatars_enabled': False,
        }

        with self.siteconfig_settings(siteconfig_settings):
            self.assertJSONEqual(
                self._render_tag(self.user),
                {
                    'allowSelfShipIt': True,
                    'archivedReviewRequestsURL':
                        '/api/users/test/archived-review-requests/',
                    'authenticated': True,
                    'avatarURLs': {},
                    'avatarHTML': {},
                    'commentsOpenAnIssue': True,
                    'confirmShipIt': True,
                    'defaultUseRichText': True,
                    'enableDesktopNotifications': True,
                    'enableSpellChecking': None,
                    'fullName': 'Test User',
                    'mutedReviewRequestsURL':
                        '/api/users/test/muted-review-requests/',
                    'quickAccessActionIDs': [],
                    'readOnly': False,
                    'sessionURL': '/api/session/',
                    'timezoneOffset': '+0000',
                    'userFileAttachmentsURL':
                        '/api/users/test/user-file-attachments/',
                    'userPageURL': '/users/test/',
                    'username': 'test',
                    'watchedReviewGroupsURL':
                        '/api/users/test/watched/review-groups/',
                    'watchedReviewRequestsURL':
                        '/api/users/test/watched/review-requests/',
                })

    def _render_tag(self, user, local_site=None):
        """Utility function to render the tag.

        This will render the template for the tag, providing all necessary
        arguments. It will also check the rendered string is of the expected
        type before returning it.

        Args:
            user (django.contrib.auth.models.User):
                The user to pass to the tag.

            local_site (reviewboard.site.models.LocalSite, optional):
                An optional Local Site for the request.

        Returns:
            django.utils.safestring.SafeText:
            The rendered content.
        """
        request = RequestFactory().get('/s/local-site-1/')
        request.user = user

        if local_site is not None:
            request._local_site_name = local_site.name
            request.local_site = local_site

        t = Template('{% load accounts %}'
                     '{% js_user_session_info %}')

        rendered = t.render(Context({
            'request': request,
        }))

        self.assertIsInstance(rendered, SafeText)

        return rendered


class UserBadgesTests(TestCase):
    """Unit tests for {% user_badges %}.

    Version Added:
        8.0
    """

    def test_with_no_badges(self) -> None:
        """Testing {% user_badges %} with no badges"""
        user = self.create_user()
        t = Template(
            '{% load accounts %}'
            '{% user_badges user %}'
        )

        rendered = t.render(Context({
            'user': user,
        }))

        self.assertIsInstance(rendered, SafeText)

        self.assertEqual(rendered, '')

    def test_with_badges(self) -> None:
        """Testing {% user_badges %} with badges"""
        class MyUserDetailsProvider(BaseUserDetailsProvider):
            user_details_provider_id = 'my-user-details-provider'

            def get_user_badges(self, user, **kwargs):
                yield UserBadge(user=user,
                                label='Badge 1')
                yield UserBadge(user=user,
                                label='Badge 2',
                                css_class='custom-badge-css')

        user = self.create_user()
        t = Template(
            '{% load accounts %}'
            '{% user_badges user %}'
        )

        provider = MyUserDetailsProvider()

        try:
            user_details_provider_registry.register(provider)

            rendered = t.render(Context({
                'user': user,
            }))
        finally:
            user_details_provider_registry.unregister(provider)

        self.assertIsInstance(rendered, SafeText)

        self.assertHTMLEqual(
            rendered,
            '<span class="rb-c-user-badges"'
            '      aria-label="Badges"'
            '      role="list">'
            ' <span class="rb-c-user-badge"'
            '       role="listitem">'
            '  Badge 1'
            ' </span>'
            ' <span class="rb-c-user-badge custom-badge-css"'
            '       role="listitem">'
            '  Badge 2'
            ' </span>'
            '</span>')

    def test_with_badges_with_crash(self) -> None:
        """Testing {% user_badges %} with crash in get_user_badges()"""
        class MyUserDetailsProvider(BaseUserDetailsProvider):
            user_details_provider_id = 'my-user-details-provider'

            def get_user_badges(self, user, **kwargs):
                yield UserBadge(user=user,
                                label='Badge 1')

                raise Exception('oh no')

        user = self.create_user()
        t = Template(
            '{% load accounts %}'
            '{% user_badges user %}'
        )

        provider = MyUserDetailsProvider()

        try:
            user_details_provider_registry.register(provider)

            with self.assertLogs() as cm:
                rendered = t.render(Context({
                    'user': user,
                }))
        finally:
            user_details_provider_registry.unregister(provider)

        self.assertEqual(len(cm.output), 1)
        self.assertRegex(
            cm.output[0],
            r'^ERROR:reviewboard\.accounts\.templatetags\.accounts:'
            r'Unexpected error when fetching user badges from provider '
            r'<reviewboard\.accounts\.tests\.test_template_tags\.'
            r'UserBadgesTests\.test_with_badges_with_crash\.<locals>\.'
            r'MyUserDetailsProvider object at 0x[a-f0-9]+>: oh no\n.*'
            r'Traceback')

        self.assertIsInstance(rendered, SafeText)

        self.assertHTMLEqual(
            rendered,
            '<span class="rb-c-user-badges"'
            '      aria-label="Badges"'
            '      role="list">'
            ' <span class="rb-c-user-badge"'
            '       role="listitem">'
            '  Badge 1'
            ' </span>'
            '</span>')


class UserProfileDisplayNameTests(TestCase):
    """Unit tests for {% user_profile_display_name %}."""

    def test_escapes_name(self):
        """Testing {% user_profile_display_name %} escapes content"""
        user = User.objects.create(
            username='test',
            first_name='"><script>alert("unsafe");</script>',
            last_name='<"',
            email='test@example.com')
        request_user = User.objects.create(username='admin',
                                           email='admin@example.com',
                                           is_superuser=True)
        rendered = self._render_tag(user, request_user)

        self.assertEqual(
            rendered,
            '&quot;&gt;&lt;script&gt;alert(&quot;unsafe&quot;);&lt;/script&gt;'
            ' &lt;&quot;')

    def _render_tag(self, user, request_user):
        """Utility function to render the tag.

        This will render the template for the tag, providing all necessary
        arguments. It will also check the rendered string is of the expected
        type before returning it.

        Args:
            user (django.contrib.auth.models.User):
                The user to pass to the tag.

            request_user (django.contrib.auth.models.User):
                The requesting user.

        Returns:
            django.utils.safestring.SafeText:
            The rendered content.
        """
        request = RequestFactory().get('/')
        request.user = request_user

        t = Template('{% load accounts %}'
                     '{% user_profile_display_name user %}')

        rendered = t.render(Context({
            'request': request,
            'user': user,
        }))

        self.assertIsInstance(rendered, SafeText)

        return rendered



"""Review UI framework.

Review UIs allow specialized review experiences for different file types.
"""

from djblets.registries.importer import lazy_import_registry


#: The registry for Review UIs.
#:
#: Version Added:
#:     7.0
review_ui_registry = lazy_import_registry(
    'reviewboard.reviews.ui.registry', 'ReviewUIRegistry')



"""Form-related classes for the administration Change Form pages."""

import itertools

from django import forms
from django.contrib.admin.helpers import Fieldline, Fieldset
from django.contrib.admin.widgets import FilteredSelectMultiple
from django.contrib.auth.models import User
from django.forms.utils import flatatt
from django.template.defaultfilters import capfirst, linebreaksbr
from django.template.loader import render_to_string
from django.utils.encoding import force_str
from django.utils.html import conditional_escape, format_html
from django.utils.safestring import mark_safe

from reviewboard.admin.form_widgets import (RelatedGroupWidget,
                                            RelatedRepositoryWidget,
                                            RelatedUserWidget)
from reviewboard.reviews.models import Group
from reviewboard.scmtools.models import Repository


class ChangeFormFieldset(Fieldset):
    """A fieldset in an administration change form.

    This takes care of providing state to the change form to represent a
    fieldset and each row in that fieldset.

    The fieldset makes use of the ``.rb-c-form-fieldset`` CSS component.
    """

    def __init__(self, form, classes=(), **kwargs):
        """Initialize the fieldset.

        Args:
            form (django.contrib.admin.helpers.AdminForm):
                The administration form owning the fieldset.

            classes (tuple, optional):
                Additional CSS classes to add to the ``<fieldset>`` element.

            **kwargs (dict):
                Keyword arguments to pass to the parent class.
        """
        if classes:
            # Transform any Django-named CSS classes to what we expect for our
            # rb-c-admin-fieldset CSS component's modifiers.
            css_class_map = {
                'collapse': ('-can-collapse', '-is-collapsed'),
                'wide': ('-is-wide',),
            }

            classes = tuple(itertools.chain.from_iterable(
                css_class_map.get(css_class, (css_class,))
                for css_class in classes
            ))

        self.collapsed = '-is-collapsed' in classes

        super(ChangeFormFieldset, self).__init__(
            form,
            classes=('rb-c-form-fieldset',) + classes,
            **kwargs)

    def render(self, context):
        """Render the fieldset to HTML.

        This will default to rendering using the
        ``admin/includes/fieldset.html`` template. A
        :py:class:`~django.contrib.admin.ModelAdmin` subclass my define a
        ``fieldset_template_name`` attribute specifying an alternative template
        to use for its fieldsets.

        The template will inherit the provided context, and will contain
        this fieldset instance as ``fieldset``.

        Args:
            context (django.template.Context):
                The current template context.

        Returns:
            django.utils.safestring.SafeText:
            The resulting HTML for the fieldset.
        """
        template_name = (
            getattr(self.model_admin, 'fieldset_template_name', None) or
            'admin/includes/fieldset.html'
        )

        with context.push():
            context['fieldset'] = self

            return render_to_string(template_name, context.flatten())

    def __iter__(self):
        """Iterate through the rows of the fieldset.

        Yields:
            ChangeFormRow:
            A row in the fieldset.
        """
        readonly_fields = self.readonly_fields
        model_admin = self.model_admin

        for field in self.fields:
            yield ChangeFormRow(form=self.form,
                                field=field,
                                readonly_fields=readonly_fields,
                                model_admin=model_admin)


class ChangeFormRow(Fieldline):
    """A row in a fieldset containing one or more fields.

    A row may contain multiple fields (though it usually contains only one).

    This makes use of the ``.rb-c-form-row`` CSS component.
    """

    def __init__(self, *args, **kwargs):
        """Initialize the row.

        Args:
            *args (tuple):
                Positional arguments to pass to the parent class.

            **kwargs (dict):
                Keyword arguments to pass to the parent class.
        """
        super(ChangeFormRow, self).__init__(*args, **kwargs)

        self.is_multi_line = len(self.fields) > 1
        self.classes = ' '.join(['rb-c-form-row'] + [
            'field-%s' % field_name
            for field_name in self.fields
        ])
        self.row_id = 'row-%s' % self.fields[0]

    def __iter__(self):
        """Iterate through the list of fields in the row.

        Yields:
            ChangeFormField:
            A field in the row.
        """
        for admin_field in super(ChangeFormRow, self).__iter__():
            yield ChangeFormField(self, admin_field)


class ChangeFormField(object):
    """A wrapper for a field on the change form.

    This takes care of providing state to the change form to represent an
    individual field on a row, providing any field validation errors.

    It also takes care of creating ideal representations of some widgets
    (such as our special related object widgets for users, groups, and
    repositories, and filtered multi-select for other many-to-many relations).

    This makes use of the ``.rb-c-form-field`` CSS component.
    """

    def __init__(self, form_row, admin_field):
        """Initialize the field wrapper.

        Args:
            form_row (ChangeFormRow):
                The parent row containing the field.

            admin_field (django.contrib.admin.helpers.AdminField):
                The administration field wrapper containing state for this
                field.
        """
        bound_field = admin_field.field
        has_field_first = False
        show_errors = False
        is_checkbox = getattr(admin_field, 'is_checkbox', False)
        is_readonly = getattr(admin_field, 'is_readonly', False)

        classes = ['rb-c-form-field']

        if is_readonly:
            classes.append('-is-read-only')
            errors = []
        else:
            form_field = bound_field.field
            errors = admin_field.errors()

            if form_field.required:
                classes.append('-is-required')

            if errors:
                classes.append('-has-errors')
                show_errors = True

            if isinstance(form_field, forms.ModelMultipleChoiceField):
                widget = form_field.widget
                model = form_field.queryset.model

                if type(widget) is forms.ModelMultipleChoiceField.widget:
                    # This is a default widget for a model multi-choice field.
                    # Let's see if we use a better default.
                    if model is User:
                        form_field.widget = RelatedUserWidget()
                    elif model is Group:
                        form_field.widget = RelatedGroupWidget()
                    elif model is Repository:
                        form_field.widget = RelatedRepositoryWidget()
                    else:
                        # We can at least use the filtered selector.
                        form_field.widget = FilteredSelectMultiple(
                            form_field.label,
                            is_stacked=False)

                if type(widget) is not widget:
                    # We've replaced the widget, so get rid of the old bound
                    # help text while we're at it.
                    bound_field.help_text = None

        if form_row.is_multi_line:
            classes.append('field-%s' % bound_field.name)
        elif is_checkbox:
            classes.append('-has-input-first')
            has_field_first = True

        self.admin_field = admin_field
        self.classes = ' '.join(classes)
        self.errors = errors
        self.field = bound_field
        self.has_field_first = has_field_first
        self.is_checkbox = is_checkbox
        self.is_first = admin_field.is_first
        self.is_readonly = is_readonly
        self.show_errors = show_errors

    def label_tag(self):
        """Return the HTML for a label tag for this field.

        This will create a ``<label class="rb-c-form-field_label">`` element
        containing the label.

        Returns:
            django.utils.safestring.SafeText:
            The ``<label>`` tag for this field.
        """
        field = self.field
        attrs = {}
        classes = ['rb-c-form-field__label']

        if not self.is_first:
            classes.append('-is-inline')

        attrs['class'] = ' '.join(classes)

        if self.is_readonly:
            return format_html('<label{0}>{1}:</label>',
                               flatatt(attrs),
                               capfirst(force_str(field['label'])))
        else:
            if self.has_field_first:
                label_suffix = ''
            else:
                label_suffix = None

            return field.label_tag(
                contents=conditional_escape(force_str(field.label)),
                attrs=attrs,
                label_suffix=label_suffix)

    def render(self):
        """Render the field.

        This will return the rendered field as HTML, or just the field's value
        if the field is meant to be read-only.

        Returns:
            django.utils.safestring.SafeText:
            The rendered content for the field.
        """
        if self.is_readonly:
            return format_html(
                '<div class="rb-c-form-field__readonly-value">{0}</div>',
                linebreaksbr(self.admin_field.contents()))
        else:
            return mark_safe(str(self.field))



"""Forms and fields related to review requests and operations."""

from __future__ import annotations

import re
from typing import TYPE_CHECKING

from django import forms
from django.contrib.auth.models import User
from django.core.exceptions import ValidationError
from django.utils.translation import gettext, gettext_lazy as _
from djblets.forms.fields import ConditionsField

from reviewboard.admin.form_widgets import (RelatedGroupWidget,
                                            RelatedRepositoryWidget,
                                            RelatedUserWidget)
from reviewboard.diffviewer import forms as diffviewer_forms
from reviewboard.diffviewer.models import DiffSet
from reviewboard.reviews.conditions import review_request_condition_choices
from reviewboard.reviews.models import (DefaultReviewer, Group,
                                        ReviewRequestDraft, Screenshot)
from reviewboard.scmtools.models import Repository
from reviewboard.site.mixins import LocalSiteAwareModelFormMixin
from reviewboard.site.validation import (validate_repositories,
                                         validate_review_groups,
                                         validate_users)

if TYPE_CHECKING:
    from djblets.forms.fields import ConditionsFieldChoices
    from typelets.django.strings import StrOrPromise


def regex_validator(value):
    """Validates that the specified regular expression is valid."""
    try:
        re.compile(value)
    except Exception as e:
        raise ValidationError(e)


class DefaultReviewerForm(LocalSiteAwareModelFormMixin, forms.ModelForm):
    name = forms.CharField(
        label=_("Name"),
        max_length=64,
        widget=forms.TextInput(attrs={'size': '30'}))

    file_regex = forms.CharField(
        label=_("File regular expression"),
        max_length=256,
        widget=forms.TextInput(attrs={'size': '60'}),
        validators=[regex_validator],
        help_text=_('File paths are matched against this regular expression '
                    'to determine if these reviewers should be added.'))

    people = forms.ModelMultipleChoiceField(
        queryset=User.objects.filter(is_active=True),
        label=_('Default users'),
        required=False,
        widget=RelatedUserWidget(),
        error_messages={
            'invalid_choice': _('A user with ID %(value)s was not found.'),
            'invalid_pk_value': _('"%(pk)s" is an invalid user ID.'),
        })

    repository = forms.ModelMultipleChoiceField(
        label=_('Repositories'),
        required=False,
        queryset=Repository.objects.order_by('name'),
        help_text=_('The list of repositories to specifically match this '
                    'default reviewer for. If left empty, this will match '
                    'all repositories.'),
        widget=RelatedRepositoryWidget(),
        error_messages={
            'invalid_choice': _('A repository with ID %(value)s was not '
                                'found.'),
            'invalid_pk_value': _('"%(pk)s" is an invalid repository ID.'),
        })

    groups = forms.ModelMultipleChoiceField(
        label=_('Default groups'),
        required=False,
        queryset=Group.objects.order_by('name'),
        widget=RelatedGroupWidget())

    def clean(self):
        try:
            validate_users(self, 'people')
        except ValidationError as e:
            self._errors['people'] = self.error_class(e.messages)

        try:
            validate_review_groups(self, 'groups')
        except ValidationError as e:
            self._errors['groups'] = self.error_class(e.messages)

        try:
            validate_repositories(self, 'repository')
        except ValidationError as e:
            self._errors['repository'] = self.error_class(e.messages)

        return super(DefaultReviewerForm, self).clean()

    class Meta:
        model = DefaultReviewer
        error_messages = {
            'groups': {
                'invalid_choice': _('A group with ID %(value)s was not '
                                    'found.'),
                'invalid_pk_value': _('"%(pk)s" is an invalid group ID.'),
            },
        }
        fields = '__all__'


class GroupForm(LocalSiteAwareModelFormMixin, forms.ModelForm):
    users = forms.ModelMultipleChoiceField(
        queryset=User.objects.all(),
        label=_('Users'),
        required=False,
        widget=RelatedUserWidget(),
        error_messages={
            'invalid_choice': _('A user with ID %(value)s was not found.'),
            'invalid_pk_value': _('"%(pk)s" is an invalid user ID.'),
        })

    class Meta:
        model = Group
        fields = '__all__'


class UploadCommitForm(diffviewer_forms.UploadCommitForm):
    """A specialized UploadCommitForm for interacting with review requests."""

    def __init__(self, review_request, *args, **kwargs):
        """Initialize the form.

        Args:
            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The review request that the uploaded commit will be attached
                to.

            *args (tuple):
                Additional positional arguments.

            **kwargs (dict):
                Additional keyword arguments.
        """
        super(UploadCommitForm, self).__init__(*args, **kwargs)

        self.review_request = review_request

    def clean(self):
        """Clean the form.

        Returns:
            dict:
            The cleaned form data.

        Raises:
            django.core.exceptions.ValidationError:
                The form failed validation.
        """
        super(UploadCommitForm, self).clean()

        if not self.review_request.created_with_history:
            raise ValidationError(
                'This review request was created without commit history '
                'support.')

        return self.cleaned_data


class UploadDiffForm(diffviewer_forms.UploadDiffForm):
    """A specialized UploadDiffForm for interacting with review requests."""

    def __init__(self, review_request, request=None, *args, **kwargs):
        """Initialize the form.

        Args:
            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request that the uploaded diff will be attached to.

            request (django.http.HttpRequest):
                The current HTTP request.

            *args (tuple):
                Additional positional arguments.

            **kwargs (dict):
                Additional keyword arguments.
        """
        super(UploadDiffForm, self).__init__(
            repository=review_request.repository,
            request=request,
            *args,
            **kwargs)
        self.review_request = review_request

        if ('basedir' in self.fields and
            (not self.data or 'basedir' not in self.data)):
            try:
                diffset = DiffSet.objects.filter(
                    history=review_request.diffset_history_id).latest()
                self.fields['basedir'].initial = diffset.basedir
            except DiffSet.DoesNotExist:
                pass

    def clean(self):
        """Clean the form.

        This ensures that the associated review request was not created with
        history.

        Returns:
            dict:
            The cleaned form data.

        Raises:
            django.core.exceptions.ValidationError:
                The form cannot be validated.
        """
        super(UploadDiffForm, self).clean()

        if self.review_request.created_with_history:
            raise ValidationError(gettext(
                'The review request was created with history support and '
                'DiffSets cannot be attached in this way. Instead, attach '
                'DiffCommits.'))

        return self.cleaned_data

    def create(self, attach_to_history=False):
        """Create the DiffSet and optionally attach it to the history.

        Args:
            attach_to_history (bool):
                Whether or not the created
                :py:class:`~reviewboard.diffviewer.models.diffset.DiffSet` will
                be attached to the diffset history of the
                :py:class:`reviewboard.reviews.models.review_request.
                ReviewRequest`.

                Defaults to ``False``.

        Returns:
            reviewboard.diffviewer.models.diffset.DiffSet:
            The created DiffSet.
        """
        assert self.is_valid()
        history = None

        if attach_to_history:
            history = self.review_request.diffset_history

        diffset = super(UploadDiffForm, self).create(history)

        if not attach_to_history:
            # Set the initial revision to be one newer than the most recent
            # public revision, so we can reference it in the diff viewer.
            diffset.update_revision_from_history(
                self.review_request.diffset_history)
            diffset.save()

        return diffset


class UploadScreenshotForm(forms.Form):
    """
    A form that handles uploading of new screenshots.
    A screenshot takes a path argument and optionally a caption.
    """
    caption = forms.CharField(required=False)
    path = forms.ImageField(required=True)

    def create(self, file, review_request):
        screenshot = Screenshot(caption='',
                                draft_caption=self.cleaned_data['caption'])
        screenshot.image.save(file.name, file, save=True)

        draft = ReviewRequestDraft.create(review_request)
        draft.screenshots.add(screenshot)
        draft.save()

        return screenshot


class ReviewRequestConditionsField(ConditionsField):
    """A field for selecting review request conditions.

    This will include all conditions that are built into Review Board and
    those registered by extensions.

    Version Added:
        8.0
    """

    def __init__(
        self,
        *,
        choices: ConditionsFieldChoices = review_request_condition_choices,
        label: StrOrPromise = _('Conditions'),
        **kwargs,
    ) -> None:
        """Initialize the field.

        Args:
            choices (djblets.conditions.choices.ConditionChoices or type or
                     callable):
                The registry of choices available for the conditions.

                This defaults to the standard review request condition choices.
                It can be overridden to provide a more specialized list of
                conditions.

            label (str, optional):
                A custom label to use for this field.

            **kwargs (dict):
                Keyword arguments to pass to the parent class.
        """
        super().__init__(choices=choices,
                         label=label,
                         **kwargs)



"""Django AdminSite customization for Review Board."""

from django.contrib import admin
from django.contrib.admin.forms import AdminAuthenticationForm
from django.contrib.admin.sites import (AdminSite as DjangoAdminSite,
                                        site as _django_site)
from django.utils.translation import gettext_lazy as _

from reviewboard.accounts.forms.auth import ReviewBoardAuthenticationFormMixin


class AuthenticationForm(ReviewBoardAuthenticationFormMixin,
                         AdminAuthenticationForm):
    """Authentication form for the administration UI.

    This builds upon the main admin authentication form (which handles
    permission checks when verifying a username) by incorporating the
    capabilities of the main Review Board administration form, allowing
    e-mail addresses as usernames and rate limiting login attempts.

    Version Added:
        4.0
    """


class AdminSite(DjangoAdminSite):
    """Main administration site for Review Board.

    This provides customization of the administration UI, while maintaining
    Model/Admin registration compatibility with the standard Django
    administration UI.

    Technically, this is focused on the Database section of the administration
    UI, though it's planned to be used to consolidate more of the UI in the
    future.

    All code should reference :py:data:`admin_site` where possible, rather than
    :py:data:`django.contrib.admin.site`.

    Version Added:
        4.0
    """

    site_title = _('Database')
    site_header = _('Database')
    index_title = _('Database')

    login_form = AuthenticationForm

    def __init__(self, *args, **kwargs):
        """Initialize the site.

        Args:
            *args (tuple):
                Positional arguments for the parent class.

            **kwargs (dict):
                Keyword arguments for the parent class.
        """
        super(AdminSite, self).__init__(*args, **kwargs)

        # Perform an import for anything that's been registered up to this
        # point.
        #
        # The module will be setting this instance as the new site immediately
        # after initialization, so this only needs to be done once.
        self._registry.update(_django_site._registry)
        self._actions.update(_django_site._actions)
        self._global_actions.update(_django_site._global_actions)

    def get_model_admin(self, model_cls):
        """Return the ModelAdmin for a given Model class.

        Args:
            model_cls (type):
                The registered model class.

        Returns:
            django.contrib.admin.ModelAdmin:
            The ModelAdmin for the Model, or ``None`` if one is not registered.
        """
        return self._registry.get(model_cls)


#: The main instance for the Review Board administration site.
#:
#: Version Added:
#:     4.0
admin_site = AdminSite()


# Set our new site as the main default site for Django.
admin.site = admin_site



"""License providers common to unit tests.

Version Added:
    8.0
"""

from __future__ import annotations

from datetime import timedelta
from typing import TYPE_CHECKING

from django.utils import timezone

from reviewboard.licensing.errors import LicenseActionError
from reviewboard.licensing.license import LicenseInfo, LicenseStatus
from reviewboard.licensing.license_checks import (
    ProcessCheckLicenseResult,
    ProcessCheckLicenseResultStatus,
)
from reviewboard.licensing.provider import BaseLicenseProvider

if TYPE_CHECKING:
    from collections.abc import Sequence

    from django.http import HttpRequest
    from djblets.util.typing import JSONValue, SerializableJSONDict

    from reviewboard.licensing.license_checks import RequestCheckLicenseResult
    from reviewboard.licensing.provider import (LicenseAction,
                                                LicenseActionData)


class BasicTestsLicenseProvider(BaseLicenseProvider):
    """Sample license provider used for unit tests.

    This provides the following licenses:

    * ``license1``

      Unlicensed, expiring in 100 days.

      Includes line items.

    * ``license2``

      Licensed, expiring in 5 days.

      Includes a plan ID and name.

    This also provides a custom ``test`` action and license updating.

    Version Added:
        8.0
    """

    license_provider_id = 'basic-tests-provider'

    custom_actions = {
        'test',
    }

    def get_license_actions(
        self,
        *,
        license_info: LicenseInfo,
        request: (HttpRequest | None) = None,
    ) -> Sequence[LicenseAction]:
        """Return actions to display when viewing a license.

        This will return a custom "test" action.

        Args:
            license_info (reviewboard.licensing.license.LicenseInfo):
                The license information that will be displayed.

            request (django.http.HttpRequest, optional):
                The HTTP request from the client.

        Returns:
            list of LicenseAction:
            The list of actions to display for the license.
        """
        return [{
            'action_id': 'test',
            'label': 'Test',
            'url': f'https://example.com/{license_info.license_id}/',
        }]

    def get_licenses(self) -> Sequence[LicenseInfo]:
        """Return the list of available licenses.

        Returns:
            list of reviewboard.licensing.license.LicenseInfo:
            The list of licenses provided.
        """
        license1 = self.get_license_by_id('license1')
        license2 = self.get_license_by_id('license2')

        assert license1
        assert license2

        return [license1, license2]

    def get_license_by_id(
        self,
        license_id: str,
    ) -> LicenseInfo | None:
        """Return a license with the given provider-unique ID.

        Args:
            license_id (str):
                The provider-unique ID for the license.

        Returns:
            reviewboard.licensing.license.LicenseInfo:
            The license information, or ``None`` if not found.
        """
        if license_id == 'license1':
            return LicenseInfo(
                expires=timezone.now() + timedelta(days=100),
                license_id=license_id,
                licensed_to='Test User',
                line_items=[
                    {
                        'content': 'Line 1',
                        'icon': 'ink-i-info',
                    },
                    {
                        'content': '<strong>Line 2</strong>',
                        'content_is_html': True,
                    },
                ],
                product_name='Test Product',
                status=LicenseStatus.UNLICENSED)
        elif license_id == 'license2':
            return LicenseInfo(
                expires=timezone.now() + timedelta(days=5),
                license_id=license_id,
                licensed_to='Test User',
                plan_id='plan1',
                plan_name='Plan 1',
                product_name='Test Product',
                status=LicenseStatus.LICENSED)
        else:
            return None

    def get_manage_license_url(
        self,
        *,
        license_info: LicenseInfo,
    ) -> str | None:
        """Return the URL for managing licenses online.

        Args:
            license_info (reviewboard.licensing.license.LicenseInfo):
                The license information to link to in the portal.

        Returns:
            str:
            The URL to the license portal, or ``None`` if one is not
            available.
        """
        if license_info.license_id == 'license1':
            return f'https://example.com/{license_info.license_id}/'

        return None

    def get_check_license_request(
        self,
        *,
        action_data: LicenseActionData,
        license_info: LicenseInfo,
        request: (HttpRequest | None) = None,
    ) -> RequestCheckLicenseResult | None:
        """Return data used to check a license for validity.

        Args:
            action_data (dict):
                Data provided in the request to the action.

                This will correspond to HTTP POST data if processing via an
                HTTP request from the client.

            license_info (reviewboard.licensing.license.LicenseInfo):
                The license information that will be checked.

            request (django.http.HttpRequest, optional):
                The HTTP request from the client, if any.

        Returns:
            reviewboard.licensing.license_checks.RequestCheckLicenseResult:
            The state used to control the license check, or ``None`` if checks
            are not supported.

        Raises:
            reviewboard.licensing.errors.LicenseActionError:
                There was an error generating data for this request.
        """
        return {
            'data': {
                'license_id': license_info.license_id,
                'something': 'special',
                'version': '1.0',
            },
            'session_token': 'abc123',
            'url': f'https://example.com/{license_info.license_id}/check/'
        }

    def process_check_license_result(
        self,
        *,
        action_data: LicenseActionData,
        license_info: LicenseInfo,
        check_request_data: JSONValue,
        check_response_data: JSONValue,
        session_token: str | None,
        request: (HttpRequest | None) = None,
    ) -> ProcessCheckLicenseResult:
        """Process the result of a license check.

        Args:
            action_data (dict):
                Data provided in the request to the action.

                This will correspond to HTTP POST data if processing via an
                HTTP request from the client.

            license_info (reviewboard.licensing.license.LicenseInfo):
                The license information to convert to model data.

            check_request_data (djblets.util.typing.JSONValue):
                The original request data sent to the license check.

            check_response_data (djblets.util.typing.JSONValue):
                The response data from the license check.

            session_token (str):
                The session token provided by
                :py:meth:`get_check_license_request`.

                This will be ``None`` if a session token was not provided.

            request (django.http.HttpRequest, optional):
                The HTTP request from the client.

        Returns:
            reviewboard.licensing.license_checks.ProcessCheckLicenseResult:
            The processed result of the license check.

        Raises:
            NotImplementedError:
                This method must be implemented by subclasses.

            reviewboard.licensing.errors.LicenseActionError:
                There was an error processing license data, or the license
                data was not valid for the product.
        """
        assert isinstance(check_response_data, dict)

        if check_response_data.get('version') != '1.0':
            raise LicenseActionError(
                'Invalid version from the licensing server! Oh no!'
            )

        status: ProcessCheckLicenseResultStatus

        if check_response_data.get('updated'):
            status = ProcessCheckLicenseResultStatus.APPLIED
            license_info.expires = timezone.now() + timedelta(days=365)
            license_info.plan_id = 'smpbpe1'
            license_info.plan_name = 'Super Mega Power Bundle Pro Enterprise'
        elif check_response_data.get('latest'):
            status = ProcessCheckLicenseResultStatus.HAS_LATEST
        else:
            status = ProcessCheckLicenseResultStatus.ERROR_APPLYING

        return {
            'license_infos': {
                'license1': self.get_js_license_model_data(
                    license_info=license_info),
            },
            'status': status,
        }

    def handle_test_action(
        self,
        *,
        action_data: LicenseActionData,
        license_info: LicenseInfo,
        request: HttpRequest | None,
    ) -> SerializableJSONDict:
        """Handle a "test" action.

        This will just send a sample payload back to the caller.

        Args:
            action_data (dict):
                Data provided in the request to the action.

            license_info (reviewboard.licensing.license.LicenseInfo):
                Information on the license to check for updates.

            request (django.http.HttpRequest, optional):
                The HTTP request from the client.

        Returns:
            dict:
            The JSON-serializable dictionary of results to send back to the
            client.

        Raises:
            reviewboard.licensing.errors.LicenseActionError:
                An error with the data or with invoking the upload in the
                License Provider. This will result in a suitable error
                message for the client.
        """
        return {
            'got_action_data': action_data,
            'ok': 'yep',
        }


class ExpiredTestsLicenseProvider(BaseLicenseProvider):
    """Sample license provider used for expiration unit tests.

    This provides the following licenses:

    * ``license1``

      Hard-expired 50 days ago.

    * ``license2``

      Expired, in grace period as of 2 days ago.

    Version Added:
        8.0
    """

    license_provider_id = 'expired-tests-provider'

    def get_licenses(self) -> Sequence[LicenseInfo]:
        """Return the list of available licenses.

        Returns:
            list of reviewboard.licensing.license.LicenseInfo:
            The list of licenses provided.
        """
        license1 = self.get_license_by_id('license1')
        license2 = self.get_license_by_id('license2')

        assert license1
        assert license2

        return [license1, license2]

    def get_license_by_id(
        self,
        license_id: str,
    ) -> LicenseInfo | None:
        """Return a license with the given provider-unique ID.

        Args:
            license_id (str):
                The provider-unique ID for the license.

        Returns:
            reviewboard.licensing.license.LicenseInfo:
            The license information, or ``None`` if not found.
        """
        if license_id == 'license1':
            return LicenseInfo(
                expires=timezone.now() - timedelta(days=50),
                is_trial=True,
                license_id=license_id,
                licensed_to='Test User',
                product_name='Test Product',
                status=LicenseStatus.HARD_EXPIRED)
        elif license_id == 'license2':
            return LicenseInfo(
                expires=timezone.now() - timedelta(days=2),
                grace_period_days_remaining=3,
                license_id=license_id,
                licensed_to='Test User',
                product_name='Test Product',
                status=LicenseStatus.EXPIRED_GRACE_PERIOD)
        else:
            return None
